"""Various classes that perform a transformation on instance sets."""

from __future__ import annotations

import logging
from collections import OrderedDict
from dataclasses import dataclass
from itertools import chain
from typing import Dict, List, Optional, Sequence, Tuple

from dbt_semantic_interfaces.references import EntityReference, MetricReference, SemanticModelReference
from dbt_semantic_interfaces.type_enums.aggregation_type import AggregationType
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity
from metricflow_semantics.aggregation_properties import AggregationState
from metricflow_semantics.assert_one_arg import assert_exactly_one_arg_set
from metricflow_semantics.instances import (
    DimensionInstance,
    EntityInstance,
    GroupByMetricInstance,
    InstanceSet,
    InstanceSetTransform,
    MdoInstance,
    MeasureInstance,
    MetadataInstance,
    MetricInstance,
    TimeDimensionInstance,
)
from metricflow_semantics.model.semantics.metric_lookup import MetricLookup
from metricflow_semantics.model.semantics.semantic_model_lookup import SemanticModelLookup
from metricflow_semantics.specs.column_assoc import ColumnAssociationResolver
from metricflow_semantics.specs.instance_spec import InstanceSpec, LinkableInstanceSpec
from metricflow_semantics.specs.measure_spec import MeasureSpec, MetricInputMeasureSpec
from metricflow_semantics.specs.spec_set import InstanceSpecSet
from metricflow_semantics.sql.sql_exprs import (
    SqlAggregateFunctionExpression,
    SqlColumnReference,
    SqlColumnReferenceExpression,
    SqlExpressionNode,
    SqlFunction,
    SqlFunctionExpression,
    SqlStringExpression,
)
from more_itertools import bucket

from metricflow.dataflow.nodes.join_to_base import ValidityWindowJoinDescription
from metricflow.plan_conversion.select_column_gen import SelectColumnSet
from metricflow.sql.sql_plan import (
    SqlSelectColumn,
)

logger = logging.getLogger(__name__)


class CreateSelectColumnsForInstances(InstanceSetTransform[SelectColumnSet]):
    """Create select column expressions that will express all instances in the set.

    It assumes that the column names of the instances are represented by the supplied column association resolver and
    come from the given table alias.
    """

    def __init__(
        self,
        table_alias: str,
        column_resolver: ColumnAssociationResolver,
        output_to_input_column_mapping: Optional[OrderedDict[str, str]] = None,
    ) -> None:
        """Initializer.

        Args:
            table_alias: the table alias to select columns from
            column_resolver: resolver to name columns.
            output_to_input_column_mapping: if specified, use these columns in the input for the given output columns.
        """
        self._table_alias = table_alias
        self._column_resolver = column_resolver
        self._output_to_input_column_mapping = output_to_input_column_mapping or OrderedDict()

    def transform(self, instance_set: InstanceSet) -> SelectColumnSet:  # noqa: D102
        metric_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.metric_instances])
        )
        measure_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.measure_instances])
        )
        dimension_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.dimension_instances])
        )
        time_dimension_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.time_dimension_instances])
        )
        entity_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.entity_instances])
        )
        metadata_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.metadata_instances])
        )
        group_by_metric_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.group_by_metric_instances])
        )
        return SelectColumnSet(
            metric_columns=metric_cols,
            measure_columns=measure_cols,
            dimension_columns=dimension_cols,
            time_dimension_columns=time_dimension_cols,
            entity_columns=entity_cols,
            group_by_metric_columns=group_by_metric_cols,
            metadata_columns=metadata_cols,
        )

    def _make_sql_column_expression(
        self,
        element_instance: MdoInstance,
    ) -> List[SqlSelectColumn]:
        """Convert one element instance into a SQL column."""
        # Do a sanity check to make sure that there's a 1:1 mapping between the columns associations generated by the
        # column resolver based on the spec, and the columns that are already associated with the instance.
        expected_column_associations = (self._column_resolver.resolve_spec(element_instance.spec),)
        existing_column_associations = element_instance.associated_columns

        # Dict between the expected column name and the corresponding column in the existing columns
        column_matches: Dict[str, List[str]] = {
            expected_column.column_name: [
                col.column_name
                for col in existing_column_associations
                if col.column_correlation_key == expected_column.column_correlation_key
            ]
            for expected_column in expected_column_associations
        }

        # Assert a 1:1 mapping between expected and existing
        assert all([len(x) == 1 for x in column_matches.values()]), (
            f"Did not find exactly one match for each expected column associations.  "
            f"Expected -> existing mappings: {column_matches}"
        )
        existing_names = set([col.column_name for col in existing_column_associations])
        mapped_names = set()
        mapped_cols: List[str] = []
        for mapped_cols in column_matches.values():
            mapped_names.update([col_name for col_name in mapped_cols])
        assert existing_names == mapped_names, (
            f"Not all existing columns were mapped. Existing: {existing_names}.  Mapped: {mapped_cols}, "
            f"{expected_column_associations} -- {existing_column_associations}"
        )

        select_columns = []
        for expected_name, mapped_cols in column_matches.items():
            input_column_name = mapped_cols[0]
            output_column_name = expected_name

            if output_column_name in self._output_to_input_column_mapping:
                input_column_name = self._output_to_input_column_mapping[output_column_name]
            select_columns.append(
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression.create(SqlColumnReference(self._table_alias, input_column_name)),
                    column_alias=output_column_name,
                )
            )
        return select_columns


class CreateSelectColumnsWithMeasuresAggregated(CreateSelectColumnsForInstances):
    """Create select columns of the form "fct_bookings.bookings AS bookings" for all the instances.

    However, for measure columns, convert them into expressions like "SUM(fct_bookings.bookings) AS bookings" so that
    the resulting expressions can be used for aggregations.

    Also add an output alias that conforms to the alias
    """

    def __init__(  # noqa: D107
        self,
        table_alias: str,
        column_resolver: ColumnAssociationResolver,
        semantic_model_lookup: SemanticModelLookup,
        metric_input_measure_specs: Sequence[MetricInputMeasureSpec],
    ) -> None:
        self._semantic_model_lookup = semantic_model_lookup
        self.metric_input_measure_specs = metric_input_measure_specs
        super().__init__(table_alias=table_alias, column_resolver=column_resolver)

    def _make_sql_column_expression_to_aggregate_measures(
        self, measure_instances: Tuple[MeasureInstance, ...]
    ) -> List[SqlSelectColumn]:
        output_columns: List[SqlSelectColumn] = []
        aliased_input_specs = [spec for spec in self.metric_input_measure_specs if spec.alias]
        for instance in measure_instances:
            matches = [spec for spec in aliased_input_specs if spec.measure_spec == instance.spec]
            if matches:
                aliased_spec = matches[0]
                aliased_input_specs.remove(aliased_spec)
                output_measure_spec = aliased_spec.post_aggregation_spec
            else:
                output_measure_spec = instance.spec

            output_columns.append(
                self._make_sql_column_expression_to_aggregate_measure(
                    measure_instance=instance, output_measure_spec=output_measure_spec
                )
            )

        return output_columns

    def _make_sql_column_expression_to_aggregate_measure(
        self, measure_instance: MeasureInstance, output_measure_spec: MeasureSpec
    ) -> SqlSelectColumn:
        """Convert one measure instance into a SQL column."""
        # Get the column name of the measure in the table that we're reading from
        column_name_in_table = measure_instance.associated_column.column_name

        # Create an expression that will aggregate the given measure.
        # Figure out the aggregation function for the measure.
        measure = self._semantic_model_lookup.measure_lookup.get_measure(measure_instance.spec.reference)
        aggregation_type = measure.agg

        expression_to_get_measure = SqlColumnReferenceExpression.create(
            SqlColumnReference(self._table_alias, column_name_in_table)
        )

        expression_to_aggregate_measure = SqlFunctionExpression.build_expression_from_aggregation_type(
            aggregation_type=aggregation_type,
            sql_column_expression=expression_to_get_measure,
            agg_params=measure.agg_params,
        )

        # Get the output column name from the measure/alias

        new_column_association_for_aggregated_measure = self._column_resolver.resolve_spec(output_measure_spec)
        new_column_name_for_aggregated_measure = new_column_association_for_aggregated_measure.column_name

        return SqlSelectColumn(
            expr=expression_to_aggregate_measure,
            column_alias=new_column_name_for_aggregated_measure,
        )

    def transform(self, instance_set: InstanceSet) -> SelectColumnSet:  # noqa: D102
        metric_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.metric_instances])
        )

        measure_cols = self._make_sql_column_expression_to_aggregate_measures(instance_set.measure_instances)
        dimension_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.dimension_instances])
        )
        time_dimension_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.time_dimension_instances])
        )
        entity_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.entity_instances])
        )
        metadata_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.metadata_instances])
        )
        group_by_metric_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.group_by_metric_instances])
        )
        return SelectColumnSet(
            metric_columns=metric_cols,
            measure_columns=measure_cols,
            dimension_columns=dimension_cols,
            time_dimension_columns=time_dimension_cols,
            entity_columns=entity_cols,
            group_by_metric_columns=group_by_metric_cols,
            metadata_columns=metadata_cols,
        )


@dataclass(frozen=True)
class _DimensionValidityParams:
    """Helper dataclass for managing dimension validity properties."""

    dimension_name: str
    time_granularity: TimeGranularity
    date_part: Optional[DatePart] = None


class CreateValidityWindowJoinDescription(InstanceSetTransform[Optional[ValidityWindowJoinDescription]]):
    """Create and return a ValidityWindowJoinDescription based on the given InstanceSet.

    During join resolution we need to determine whether or not a given data set represents a
    Type II SCD dataset - i.e., one with a validity window defined on each row. This requires
    checking the set of dimension instances and determining whether or not those originate from
    an SCD source, and extracting validity window information accordingly.
    """

    def __init__(self, semantic_model_lookup: SemanticModelLookup) -> None:
        """Initializer. The SemanticModelLookup is needed for getting the original model definition."""
        self._semantic_model_lookup = semantic_model_lookup

    def _get_validity_window_dimensions_for_semantic_model(
        self, semantic_model_reference: SemanticModelReference
    ) -> Optional[Tuple[_DimensionValidityParams, _DimensionValidityParams]]:
        """Returns a 2-tuple (start, end) of validity window dimensions info, if any exist in the semantic model."""
        semantic_model = self._semantic_model_lookup.get_by_reference(semantic_model_reference)
        assert semantic_model, f"Could not find semantic model {semantic_model_reference} after data set conversion!"

        start_dim = semantic_model.validity_start_dimension
        end_dim = semantic_model.validity_end_dimension

        # We do this instead of relying on has_validity_dimensions because this also does type refinement
        if not start_dim or not end_dim:
            return None

        assert start_dim.type_params, "Typechecker hint - validity info cannot exist without type params"
        assert end_dim.type_params, "Typechecker hint - validity info cannot exist without type params"

        return (
            _DimensionValidityParams(
                dimension_name=start_dim.name, time_granularity=start_dim.type_params.time_granularity
            ),
            _DimensionValidityParams(
                dimension_name=end_dim.name, time_granularity=end_dim.type_params.time_granularity
            ),
        )

    def transform(self, instance_set: InstanceSet) -> Optional[ValidityWindowJoinDescription]:
        """Find the Time Dimension specs defining a validity window, if any, and return it.

        This currently throws an exception if more than one such window is found, and effectively prevents
        us from processing a dataset composed of a join between two SCD semantic models. This restriction is in
        place as a temporary simplification - if there is need for this feature we can enable it.
        """
        semantic_model_to_window: Dict[SemanticModelReference, ValidityWindowJoinDescription] = {}
        instances_by_semantic_model = bucket(
            instance_set.time_dimension_instances, lambda x: x.origin_semantic_model_reference.semantic_model_reference
        )
        for semantic_model_reference in instances_by_semantic_model:
            validity_dims = self._get_validity_window_dimensions_for_semantic_model(semantic_model_reference)
            if validity_dims is None:
                continue

            start_dim, end_dim = validity_dims
            specs = {instance.spec for instance in instances_by_semantic_model[semantic_model_reference]}
            start_specs = [
                spec
                for spec in specs
                if spec.element_name == start_dim.dimension_name
                # TODO: [custom_granularity] - support custom granularities for SCDs. Note this requires
                # addition of SCD support for window sub-selection, similar to what we do for cumulative metrics
                # when we group by a different time grain (e.g., select last_value from window, etc.)
                and not spec.time_granularity.is_custom_granularity
                and spec.time_granularity.base_granularity == start_dim.time_granularity
                and spec.date_part == start_dim.date_part
            ]
            end_specs = [
                spec
                for spec in specs
                if spec.element_name == end_dim.dimension_name
                # TODO: [custom_granularity] - support custom granularities for SCDs. Note this requires
                # addition of SCD support for window sub-selection, similar to what we do for cumulative metrics
                # when we group by a different time grain (e.g., select last_value from window, etc.)
                and not spec.time_granularity.is_custom_granularity
                and spec.time_granularity.base_granularity == end_dim.time_granularity
                and spec.date_part == end_dim.date_part
            ]
            linkless_start_specs = {spec.without_entity_links for spec in start_specs}
            linkless_end_specs = {spec.without_entity_links for spec in end_specs}
            assert len(linkless_start_specs) == 1 and len(linkless_end_specs) == 1, (
                f"Did not find exactly one pair of specs from semantic model `{semantic_model_reference}` matching the validity "
                f"window end points defined in the semantic model. This means we cannot process an SCD join, because we "
                f"require exactly one validity window to be specified for the query! The window in the semantic model "
                f"is defined by start dimension `{start_dim}` and end dimension `{end_dim}`. We found "
                f"{len(linkless_start_specs)} linkless specs for window start ({linkless_start_specs}) and "
                f"{len(linkless_end_specs)} linkless specs for window end ({linkless_end_specs})."
            )
            # SCD join targets are joined as dimension links in much the same was as partitions are joined. Therefore,
            # we treat this like a partition time column join and take the dimension spec with the shortest set of
            # entity links so that the subquery uses the correct reference in the ON statement
            start_specs = sorted(start_specs, key=lambda x: len(x.entity_links))
            end_specs = sorted(end_specs, key=lambda x: len(x.entity_links))
            semantic_model_to_window[semantic_model_reference] = ValidityWindowJoinDescription(
                window_start_dimension=start_specs[0], window_end_dimension=end_specs[0]
            )

        assert len(semantic_model_to_window) < 2, (
            f"Found more than 1 set of validity window specs in the input instance set. This is not currently "
            f"supported, as joins between SCD semantic models are not yet allowed! {semantic_model_to_window}"
        )

        if semantic_model_to_window:
            return list(semantic_model_to_window.values())[0]

        return None


class FilterLinkableInstancesWithLeadingLink(InstanceSetTransform[InstanceSet]):
    """Return an instance set with the elements that have a specified leading link removed.

    e.g. Remove "listing__country" if the specified link is "listing".
    """

    def __init__(self, entity_link: EntityReference) -> None:
        """Remove elements with this link as the first element in "entity_links"."""
        self._entity_link = entity_link

    def _should_pass(self, linkable_spec: LinkableInstanceSpec) -> bool:
        if len(linkable_spec.entity_links) == 0:
            return not linkable_spec.reference == self._entity_link
        return linkable_spec.entity_links[0] != self._entity_link

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        # Normal to not filter anything if the instance set has no instances with links.
        filtered_dimension_instances = tuple(x for x in instance_set.dimension_instances if self._should_pass(x.spec))
        filtered_time_dimension_instances = tuple(
            x for x in instance_set.time_dimension_instances if self._should_pass(x.spec)
        )
        filtered_entity_instances = tuple(x for x in instance_set.entity_instances if self._should_pass(x.spec))
        filtered_group_by_metric_instances = tuple(
            x for x in instance_set.group_by_metric_instances if self._should_pass(x.spec)
        )

        output = InstanceSet(
            measure_instances=instance_set.measure_instances,
            dimension_instances=filtered_dimension_instances,
            time_dimension_instances=filtered_time_dimension_instances,
            entity_instances=filtered_entity_instances,
            group_by_metric_instances=filtered_group_by_metric_instances,
            metric_instances=instance_set.metric_instances,
            metadata_instances=instance_set.metadata_instances,
        )
        return output


class FilterElements(InstanceSetTransform[InstanceSet]):
    """Return an instance set with the elements that don't match any of the pass specs removed."""

    def __init__(
        self,
        include_specs: Optional[InstanceSpecSet] = None,
        exclude_specs: Optional[InstanceSpecSet] = None,
    ) -> None:
        """Constructor.

        Args:
            include_specs: If specified, pass only instances matching these specs.
            exclude_specs: If specified, pass only instances not matching these specs.
        """
        assert_exactly_one_arg_set(include_specs=include_specs, exclude_specs=exclude_specs)
        self._include_specs = include_specs
        self._exclude_specs = exclude_specs

    def _should_pass(self, element_spec: InstanceSpec) -> bool:
        # TODO: Use better matching function
        if self._include_specs:
            return any(x == element_spec for x in self._include_specs.all_specs)
        elif self._exclude_specs:
            return not any(x == element_spec for x in self._exclude_specs.all_specs)
        assert False

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        # Sanity check to make sure the specs are in the instance set

        if self._include_specs:
            include_specs_not_found = []
            for include_spec in self._include_specs.all_specs:
                if include_spec not in instance_set.spec_set.all_specs:
                    include_specs_not_found.append(include_spec)
            if include_specs_not_found:
                raise RuntimeError(
                    f"Include specs {include_specs_not_found} are not in the spec set {instance_set.spec_set} - "
                    f"check if this node was constructed correctly."
                )
        elif self._exclude_specs:
            exclude_specs_not_found = []
            for exclude_spec in self._exclude_specs.all_specs:
                if exclude_spec not in instance_set.spec_set.all_specs:
                    exclude_specs_not_found.append(exclude_spec)
            if exclude_specs_not_found:
                raise RuntimeError(
                    f"Exclude specs {exclude_specs_not_found} are not in the spec set {instance_set.spec_set} - "
                    f"check if this node was constructed correctly."
                )
        else:
            assert False, "Include specs or exclude specs should have been specified."

        output = InstanceSet(
            measure_instances=tuple(x for x in instance_set.measure_instances if self._should_pass(x.spec)),
            dimension_instances=tuple(x for x in instance_set.dimension_instances if self._should_pass(x.spec)),
            time_dimension_instances=tuple(
                x for x in instance_set.time_dimension_instances if self._should_pass(x.spec)
            ),
            entity_instances=tuple(x for x in instance_set.entity_instances if self._should_pass(x.spec)),
            group_by_metric_instances=tuple(
                x for x in instance_set.group_by_metric_instances if self._should_pass(x.spec)
            ),
            metric_instances=tuple(x for x in instance_set.metric_instances if self._should_pass(x.spec)),
            metadata_instances=tuple(x for x in instance_set.metadata_instances if self._should_pass(x.spec)),
        )
        return output


class ChangeMeasureAggregationState(InstanceSetTransform[InstanceSet]):
    """Returns a new instance set where all measures are set as a different aggregation state."""

    def __init__(self, aggregation_state_changes: Dict[AggregationState, AggregationState]) -> None:
        """Constructor.

        Args:
            aggregation_state_changes: key is old aggregation state, value is the new aggregation state.
        """
        self._aggregation_state_changes = aggregation_state_changes

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        for measure_instance in instance_set.measure_instances:
            assert measure_instance.aggregation_state in self._aggregation_state_changes, (
                f"Aggregation state: {measure_instance.aggregation_state} not handled in change dict: "
                f"{self._aggregation_state_changes}"
            )

        # Copy the measures, but just change the aggregation state to COMPLETE.
        measure_instances = tuple(
            MeasureInstance(
                associated_columns=x.associated_columns,
                defined_from=x.defined_from,
                aggregation_state=self._aggregation_state_changes[x.aggregation_state],
                spec=x.spec,
            )
            for x in instance_set.measure_instances
        )
        return InstanceSet(
            measure_instances=measure_instances,
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            entity_instances=instance_set.entity_instances,
            group_by_metric_instances=instance_set.group_by_metric_instances,
            metric_instances=instance_set.metric_instances,
            metadata_instances=instance_set.metadata_instances,
        )


class UpdateMeasureFillNullsWith(InstanceSetTransform[InstanceSet]):
    """Returns a new instance set where all measures have been assigned the fill nulls with property."""

    def __init__(self, metric_input_measure_specs: Sequence[MetricInputMeasureSpec]):
        """Initializer stores the input specs, which contain the fill_nulls_with for each measure."""
        self.metric_input_measure_specs = metric_input_measure_specs

    def _update_fill_nulls_with(self, measure_instances: Tuple[MeasureInstance, ...]) -> Tuple[MeasureInstance, ...]:
        """Update all measure instances with the corresponding fill_nulls_with value."""
        updated_instances: List[MeasureInstance] = []
        for instance in measure_instances:
            # There might be multiple matching specs, but they'll all have the same fill_nulls_with value so we can use any of them.
            matching_specs = [spec for spec in self.metric_input_measure_specs if spec.measure_spec == instance.spec]
            assert (
                len(matching_specs) >= 1
            ), f"Found no matching input measure spec for measure instance {instance}. Something has been misconfigured."

            measure_spec = MeasureSpec(
                element_name=instance.spec.element_name,
                fill_nulls_with=matching_specs[0].fill_nulls_with,
                non_additive_dimension_spec=instance.spec.non_additive_dimension_spec,
            )
            updated_instances.append(
                MeasureInstance(
                    associated_columns=instance.associated_columns,
                    spec=measure_spec,
                    aggregation_state=instance.aggregation_state,
                    defined_from=instance.defined_from,
                )
            )

        return tuple(updated_instances)

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        return InstanceSet(
            measure_instances=self._update_fill_nulls_with(instance_set.measure_instances),
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            entity_instances=instance_set.entity_instances,
            group_by_metric_instances=instance_set.group_by_metric_instances,
            metric_instances=instance_set.metric_instances,
            metadata_instances=instance_set.metadata_instances,
        )


class AliasAggregatedMeasures(InstanceSetTransform[InstanceSet]):
    """Returns a new instance set where all measures have been assigned an alias spec."""

    def __init__(self, metric_input_measure_specs: Sequence[MetricInputMeasureSpec]):
        """Initializer stores the input specs, which contain the aliases for each measure.

        Note this class only works if used in conjunction with an AggregateMeasuresNode that has been generated
        by querying a single semantic model for a single set of aggregated measures. This is currently enforced
        by the structure of the DataflowPlanBuilder, which ensures each AggregateMeasuresNode corresponds to
        a single semantic model set of measures for a single metric, and that these outputs will then be
        combinded via joins.
        """
        self.metric_input_measure_specs = metric_input_measure_specs

    def _alias_measure_instances(self, measure_instances: Tuple[MeasureInstance, ...]) -> Tuple[MeasureInstance, ...]:
        """Update all measure instances with aliases, if any are found in the input spec set."""
        aliased_instances: List[MeasureInstance] = []
        aliased_input_specs = [spec for spec in self.metric_input_measure_specs if spec.alias]
        for instance in measure_instances:
            matches = [spec for spec in aliased_input_specs if spec.measure_spec == instance.spec]
            assert (
                len(matches) < 2
            ), f"Found duplicate aliased measure spec matches: {matches} for measure instance {instance}. "
            "We should always have 0 or 1 matches, or else we might pass the wrong aggregated measures to the "
            "downstream metric computation expression!"
            if matches:
                aliased_spec = matches[0]
                aliased_input_specs.remove(aliased_spec)
                measure_spec = aliased_spec.post_aggregation_spec
            else:
                measure_spec = instance.spec

            aliased_instances.append(
                MeasureInstance(
                    associated_columns=instance.associated_columns,
                    spec=measure_spec,
                    aggregation_state=instance.aggregation_state,
                    defined_from=instance.defined_from,
                )
            )

        return tuple(aliased_instances)

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        return InstanceSet(
            measure_instances=self._alias_measure_instances(instance_set.measure_instances),
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            entity_instances=instance_set.entity_instances,
            group_by_metric_instances=instance_set.group_by_metric_instances,
            metric_instances=instance_set.metric_instances,
            metadata_instances=instance_set.metadata_instances,
        )


class AddMetrics(InstanceSetTransform[InstanceSet]):
    """Adds the given metric instances to the instance set."""

    def __init__(self, metric_instances: List[MetricInstance]) -> None:  # noqa: D107
        self._metric_instances = metric_instances

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        return InstanceSet(
            measure_instances=instance_set.measure_instances,
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            entity_instances=instance_set.entity_instances,
            group_by_metric_instances=instance_set.group_by_metric_instances,
            metric_instances=instance_set.metric_instances + tuple(self._metric_instances),
            metadata_instances=instance_set.metadata_instances,
        )


class AddGroupByMetric(InstanceSetTransform[InstanceSet]):
    """Adds the given metric instances to the instance set."""

    def __init__(self, group_by_metric_instance: GroupByMetricInstance) -> None:  # noqa: D107
        self._group_by_metric_instance = group_by_metric_instance

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        return InstanceSet(
            measure_instances=instance_set.measure_instances,
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            entity_instances=instance_set.entity_instances,
            group_by_metric_instances=instance_set.group_by_metric_instances + (self._group_by_metric_instance,),
            metric_instances=instance_set.metric_instances,
            metadata_instances=instance_set.metadata_instances,
        )


class RemoveMeasures(InstanceSetTransform[InstanceSet]):
    """Remove measures from the instance set."""

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        return InstanceSet(
            measure_instances=(),
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            entity_instances=instance_set.entity_instances,
            group_by_metric_instances=instance_set.group_by_metric_instances,
            metric_instances=instance_set.metric_instances,
            metadata_instances=instance_set.metadata_instances,
        )


class RemoveMetrics(InstanceSetTransform[InstanceSet]):
    """Remove metrics from the instance set."""

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        return InstanceSet(
            measure_instances=instance_set.measure_instances,
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            entity_instances=instance_set.entity_instances,
            group_by_metric_instances=instance_set.group_by_metric_instances,
            metric_instances=(),
            metadata_instances=instance_set.metadata_instances,
        )


class CreateSqlColumnReferencesForInstances(InstanceSetTransform[Tuple[SqlColumnReferenceExpression, ...]]):
    """Create select column expressions that will express all instances in the set.

    It assumes that the column names of the instances are represented by the supplied column association resolver and
    come from the given table alias.
    """

    def __init__(
        self,
        table_alias: str,
        column_resolver: ColumnAssociationResolver,
    ) -> None:
        """Initializer.

        Args:
            table_alias: the table alias to select columns from
            column_resolver: resolver to name columns.
        """
        self._table_alias = table_alias
        self._column_resolver = column_resolver

    def transform(self, instance_set: InstanceSet) -> Tuple[SqlColumnReferenceExpression, ...]:  # noqa: D102
        column_names = [
            self._column_resolver.resolve_spec(spec).column_name for spec in instance_set.spec_set.all_specs
        ]
        return tuple(
            SqlColumnReferenceExpression.create(
                SqlColumnReference(
                    table_alias=self._table_alias,
                    column_name=column_name,
                ),
            )
            for column_name in column_names
        )


class CreateSelectColumnForCombineOutputNode(InstanceSetTransform[SelectColumnSet]):
    """Create select column expressions for the instance for joining outputs.

    It assumes that the column names of the instances are represented by the supplied column association resolver and
    come from the given table alias.
    """

    def __init__(  # noqa: D107
        self,
        table_alias: str,
        column_resolver: ColumnAssociationResolver,
        metric_lookup: MetricLookup,
    ) -> None:
        self._table_alias = table_alias
        self._column_resolver = column_resolver
        self._metric_lookup = metric_lookup

    def _create_select_column(self, spec: InstanceSpec, fill_nulls_with: Optional[int] = None) -> SqlSelectColumn:
        """Creates the select column for the given spec and the fill value."""
        column_name = self._column_resolver.resolve_spec(spec).column_name
        column_reference_expression = SqlColumnReferenceExpression.create(
            col_ref=SqlColumnReference(
                table_alias=self._table_alias,
                column_name=column_name,
            )
        )
        select_expression: SqlExpressionNode = SqlFunctionExpression.build_expression_from_aggregation_type(
            aggregation_type=AggregationType.MAX, sql_column_expression=column_reference_expression
        )
        if fill_nulls_with is not None:
            select_expression = SqlAggregateFunctionExpression.create(
                sql_function=SqlFunction.COALESCE,
                sql_function_args=[
                    select_expression,
                    SqlStringExpression.create(str(fill_nulls_with)),
                ],
            )
        return SqlSelectColumn(
            expr=select_expression,
            column_alias=column_name,
        )

    def _create_select_columns_for_metrics(self, metric_instances: Tuple[MetricInstance, ...]) -> List[SqlSelectColumn]:
        select_columns: List[SqlSelectColumn] = []
        for metric_instance in metric_instances:
            metric_reference = MetricReference(element_name=metric_instance.defined_from.metric_name)
            input_measure = self._metric_lookup.configured_input_measure_for_metric(metric_reference=metric_reference)
            fill_nulls_with: Optional[int] = None
            if input_measure and input_measure.fill_nulls_with is not None:
                fill_nulls_with = input_measure.fill_nulls_with
            select_columns.append(
                self._create_select_column(spec=metric_instance.spec, fill_nulls_with=fill_nulls_with)
            )
        return select_columns

    def _create_select_columns_for_measures(
        self, measure_instances: Tuple[MeasureInstance, ...]
    ) -> List[SqlSelectColumn]:
        select_columns: List[SqlSelectColumn] = []
        for measure_instance in measure_instances:
            measure_spec = measure_instance.spec
            select_columns.append(
                self._create_select_column(spec=measure_spec, fill_nulls_with=measure_spec.fill_nulls_with)
            )
        return select_columns

    def transform(self, instance_set: InstanceSet) -> SelectColumnSet:  # noqa: D102
        return SelectColumnSet(
            metric_columns=self._create_select_columns_for_metrics(instance_set.metric_instances),
            measure_columns=self._create_select_columns_for_measures(instance_set.measure_instances),
        )


# TODO: delete this class & all uses. It doesn't do anything.
class ChangeAssociatedColumns(InstanceSetTransform[InstanceSet]):
    """Change the columns associated with instances to the one specified by the resolver.

    This is useful for conveniently generating output instances for a node that serve as a "pass-through". The output
    instances can be a copy of the parent's instances, except that the column names need to be changed.

    e.g. the parent may have a data set:
        sql:
            SELECT
                is_lux AS is_lux_latext
            ...
        instance:
            DimensionInstance(column_name="is_lux", ...)

    but for the current node, we want a data set like:

        sql:
            SELECT
                is_lux_latest
                ...
            FROM (
                -- SQL from parent
                is_lux AS is_lux_latest
                ...
            )
            ...
        instance:
            DimensionInstance(column_name="is_lux_latest")
    """

    def __init__(self, column_association_resolver: ColumnAssociationResolver) -> None:  # noqa: D107
        self._column_association_resolver = column_association_resolver

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        output_measure_instances = []
        for input_measure_instance in instance_set.measure_instances:
            output_measure_instances.append(
                MeasureInstance(
                    associated_columns=(self._column_association_resolver.resolve_spec(input_measure_instance.spec),),
                    spec=input_measure_instance.spec,
                    defined_from=input_measure_instance.defined_from,
                    aggregation_state=input_measure_instance.aggregation_state,
                )
            )

        output_dimension_instances = []
        for input_dimension_instance in instance_set.dimension_instances:
            output_dimension_instances.append(
                DimensionInstance(
                    associated_columns=(self._column_association_resolver.resolve_spec(input_dimension_instance.spec),),
                    spec=input_dimension_instance.spec,
                    defined_from=input_dimension_instance.defined_from,
                )
            )

        output_time_dimension_instances = []
        for input_time_dimension_instance in instance_set.time_dimension_instances:
            output_time_dimension_instances.append(
                TimeDimensionInstance(
                    associated_columns=(
                        self._column_association_resolver.resolve_spec(input_time_dimension_instance.spec),
                    ),
                    spec=input_time_dimension_instance.spec,
                    defined_from=input_time_dimension_instance.defined_from,
                )
            )

        output_entity_instances = []
        for input_entity_instance in instance_set.entity_instances:
            output_entity_instances.append(
                EntityInstance(
                    associated_columns=(self._column_association_resolver.resolve_spec(input_entity_instance.spec),),
                    spec=input_entity_instance.spec,
                    defined_from=input_entity_instance.defined_from,
                )
            )

        output_metric_instances = []
        for input_metric_instance in instance_set.metric_instances:
            output_metric_instances.append(
                MetricInstance(
                    associated_columns=(self._column_association_resolver.resolve_spec(input_metric_instance.spec),),
                    spec=input_metric_instance.spec,
                    defined_from=input_metric_instance.defined_from,
                )
            )

        output_metadata_instances = []
        for input_metadata_instance in instance_set.metadata_instances:
            output_metadata_instances.append(
                MetadataInstance(
                    associated_columns=(self._column_association_resolver.resolve_spec(input_metadata_instance.spec),),
                    spec=input_metadata_instance.spec,
                )
            )

        output_group_by_metric_instances = []
        for input_group_by_metric_instance in instance_set.group_by_metric_instances:
            output_group_by_metric_instances.append(
                GroupByMetricInstance(
                    associated_columns=(
                        self._column_association_resolver.resolve_spec(input_group_by_metric_instance.spec),
                    ),
                    spec=input_group_by_metric_instance.spec,
                    defined_from=input_group_by_metric_instance.defined_from,
                )
            )
        return InstanceSet(
            measure_instances=tuple(output_measure_instances),
            dimension_instances=tuple(output_dimension_instances),
            time_dimension_instances=tuple(output_time_dimension_instances),
            entity_instances=tuple(output_entity_instances),
            group_by_metric_instances=tuple(output_group_by_metric_instances),
            metric_instances=tuple(output_metric_instances),
            metadata_instances=tuple(output_metadata_instances),
        )


class ConvertToMetadata(InstanceSetTransform[InstanceSet]):
    """Removes all instances from old instance set and replaces them with a set of metadata instances."""

    def __init__(self, metadata_instances: Sequence[MetadataInstance]) -> None:  # noqa: D107
        self._metadata_instances = metadata_instances

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        return InstanceSet(
            metadata_instances=tuple(self._metadata_instances),
        )


def create_simple_select_columns_for_instance_sets(
    column_resolver: ColumnAssociationResolver,
    table_alias_to_instance_set: OrderedDict[str, InstanceSet],
) -> Tuple[SqlSelectColumn, ...]:
    """Creates select columns for instance sets coming from multiple table as defined in table_alias_to_instance_set.

    Used in cases where you join multiple tables and need to render select columns to access all of those.
    """
    column_set = SelectColumnSet()
    for table_alias, instance_set in table_alias_to_instance_set.items():
        column_set = column_set.merge(
            instance_set.transform(
                CreateSelectColumnsForInstances(
                    table_alias=table_alias,
                    column_resolver=column_resolver,
                )
            )
        )

    return column_set.as_tuple()


class AddMetadata(InstanceSetTransform[InstanceSet]):
    """Adds the given metric instances to the instance set."""

    def __init__(self, metadata_instances: Sequence[MetadataInstance]) -> None:  # noqa: D107
        self._metadata_instances = metadata_instances

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D102
        return InstanceSet(
            measure_instances=instance_set.measure_instances,
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            entity_instances=instance_set.entity_instances,
            group_by_metric_instances=instance_set.group_by_metric_instances,
            metric_instances=instance_set.metric_instances,
            metadata_instances=instance_set.metadata_instances + tuple(self._metadata_instances),
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence, Tuple

from typing_extensions import override

from metricflow_semantics.mf_logging.pretty_print import mf_pformat
from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryIssueType,
    MetricFlowQueryResolutionIssue,
)
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput
from metricflow_semantics.specs.instance_spec import LinkableInstanceSpec


@dataclass(frozen=True)
class MetricExcludesDatePartIssue(MetricFlowQueryResolutionIssue):
    """Describes an issue where the date_part can't be used in the query due to metric limitations."""

    candidate_specs: Tuple[LinkableInstanceSpec, ...]

    @staticmethod
    def from_parameters(  # noqa: D102
        parent_issues: Sequence[MetricFlowQueryResolutionIssue],
        query_resolution_path: MetricFlowQueryResolutionPath,
        candidate_specs: Sequence[LinkableInstanceSpec],
    ) -> MetricExcludesDatePartIssue:
        return MetricExcludesDatePartIssue(
            issue_type=MetricFlowQueryIssueType.ERROR,
            parent_issues=tuple(parent_issues),
            query_resolution_path=query_resolution_path,
            candidate_specs=tuple(candidate_specs),
        )

    @override
    def ui_description(self, associated_input: MetricFlowQueryResolverInput) -> str:
        # TODO: Improve error.
        lines = [
            f"{self.query_resolution_path.last_item.ui_description} does not allow group-by-items "
            f"with a date part in the query. Considered group-by-items:\n"
        ]

        if associated_input.input_pattern_description is not None:
            naming_scheme = associated_input.input_pattern_description.naming_scheme
        else:
            naming_scheme = None

        if naming_scheme is not None:
            lines.append(mf_pformat([naming_scheme.input_str(spec) for spec in self.candidate_specs]))
        else:
            lines.append(mf_pformat([spec for spec in self.candidate_specs]))

        lines.append("\nbut they were excluded by this metric.")
        return "\n".join(lines)

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> MetricExcludesDatePartIssue:
        return MetricExcludesDatePartIssue(
            issue_type=self.issue_type,
            parent_issues=tuple(issue.with_path_prefix(path_prefix) for issue in self.parent_issues),
            query_resolution_path=self.query_resolution_path.with_path_prefix(path_prefix),
            candidate_specs=self.candidate_specs,
        )



from __future__ import annotations

from enum import Enum
from typing import FrozenSet, List, Sequence

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted

from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.optimizer.dataflow_plan_optimizer import DataflowPlanOptimizer
from metricflow.dataflow.optimizer.predicate_pushdown_optimizer import PredicatePushdownOptimizer
from metricflow.dataflow.optimizer.source_scan.source_scan_optimizer import SourceScanOptimizer


class DataflowPlanOptimization(Enum):
    """Enumeration of optimization types available for execution.

    Values indicate order of application. We apply the source scan optimizer first because it reduces input branches,
    making for maximally parsimonious queries prior to application of predicate pushdown. Note this is safe only
    because the SourceScanOptimizer combines from the CombineAggregatedOutputNode, and will only combine branches
    from there to source if they are functionally identical (i.e., they have all of the same WhereConstraintNode
    configurations).
    """

    SOURCE_SCAN = 0
    PREDICATE_PUSHDOWN = 1

    @staticmethod
    def all_optimizations() -> FrozenSet[DataflowPlanOptimization]:
        """Convenience method for getting a set of all available optimizations."""
        return frozenset((DataflowPlanOptimization.SOURCE_SCAN, DataflowPlanOptimization.PREDICATE_PUSHDOWN))

    @staticmethod
    def enabled_optimizations() -> FrozenSet[DataflowPlanOptimization]:
        """Set of DataflowPlanOptimization that are currently enabled.

        Predicate pushdown optimizer is currently disabled.
        """
        return frozenset((DataflowPlanOptimization.SOURCE_SCAN,))


class DataflowPlanOptimizerFactory:
    """Factory class for initializing an enumerated set of optimizers.

    This allows us to centralize initialization and, most importantly, share class instances with cached high cost
    processing between the DataflowPlanBuilder and the optimizer instances requiring that functionality.
    """

    def __init__(self, node_data_set_resolver: DataflowPlanNodeOutputDataSetResolver) -> None:
        """Initializer.

        This collects all of the initialization requirements for the optimizers it manages.
        """
        self._node_data_set_resolver = node_data_set_resolver

    def get_optimizers(self, optimizations: FrozenSet[DataflowPlanOptimization]) -> Sequence[DataflowPlanOptimizer]:
        """Initializes and returns a sequence of optimizers matching the input optimization requests."""
        optimizers: List[DataflowPlanOptimizer] = []
        for optimization in sorted(list(optimizations), key=lambda x: x.value):
            if optimization is DataflowPlanOptimization.SOURCE_SCAN:
                optimizers.append(SourceScanOptimizer())
            elif optimization is DataflowPlanOptimization.PREDICATE_PUSHDOWN:
                optimizers.append(PredicatePushdownOptimizer(self._node_data_set_resolver))
            else:
                assert_values_exhausted(optimization)

        return tuple(optimizers)



from __future__ import annotations

import datetime
import itertools
import logging
from dataclasses import dataclass
from decimal import Decimal
from typing import Iterable, Iterator, List, Optional, Sequence, Tuple, Type

import tabulate
from metricflow_semantics.mf_logging.formatting import indent
from metricflow_semantics.mf_logging.pretty_print import mf_pformat, mf_pformat_dict
from typing_extensions import Self

from metricflow.data_table.column_types import CellValue, InputCellValue, row_cell_types
from metricflow.data_table.mf_column import ColumnDescription

logger = logging.getLogger(__name__)


@dataclass(frozen=True, eq=False)
class MetricFlowDataTable:
    """Container for tabular data stored in memory.

    This is feature-limited and is used to pass tabular data for tests and the CLI. The only data types that are
    supported are described by `CellValue`.

    When constructing the table, additional input types (as described by `InputCellValue`) can be used, but those
    additional types will be converted into one of the `CellValue` types.

    Don't use `=` to compare tables as there many be NaNs. Instead, use `check_data_tables_are_equal`.
    """

    column_descriptions: Tuple[ColumnDescription, ...]
    rows: Tuple[Tuple[CellValue, ...], ...]

    def __post_init__(self) -> None:  # noqa: D105
        expected_column_count = self.column_count
        for row_index, row in enumerate(self.rows):
            # Check that the number of columns in the rows match.
            row_column_count = len(row)
            assert row_column_count == expected_column_count, (
                f"Row at index {row_index} has {row_column_count} columns instead of {expected_column_count}. "
                f"Row is:"
                f"\n{indent(mf_pformat(row))}"
            )
            # Check that the type of the object in the rows match.
            for column_index, cell_value in enumerate(row):
                expected_cell_value_type = self.column_descriptions[column_index].column_type
                assert cell_value is None or isinstance(cell_value, expected_cell_value_type), mf_pformat_dict(
                    "Cell value type mismatch.",
                    {
                        "row_index": row_index,
                        "column_index": column_index,
                        "expected_cell_value_type": expected_cell_value_type,
                        "actual_cell_value_type": type(cell_value),
                        "cell_value": cell_value,
                    },
                )
                # Check that datetimes don't have a timezone set.
                if isinstance(cell_value, datetime.datetime):
                    assert cell_value.tzinfo is None, mf_pformat_dict(
                        "Time zone provided for datetime.",
                        {
                            "row_index": row_index,
                            "column_index": column_index,
                            "cell_value": cell_value,
                        },
                    )

    @property
    def column_count(self) -> int:  # noqa: D102
        return len(self.column_descriptions)

    @property
    def row_count(self) -> int:  # noqa: D102
        return len(self.rows)

    def column_name_index(self, column_name: str) -> int:
        """Return the index of the column that matches the given name. Raises `ValueError` if the name is invalid."""
        for i, column_description in enumerate(self.column_descriptions):
            if column_description.column_name == column_name:
                return i
        raise ValueError(
            f"Unknown column name {repr(column_name)}. Known column names are:"
            f"\n{indent(mf_pformat([column_name for column_name in self.column_descriptions]))}"
        )

    @property
    def column_names(self) -> Sequence[str]:  # noqa: D102
        return tuple(column_description.column_name for column_description in self.column_descriptions)

    def column_values_iterator(self, column_index: int) -> Iterator[CellValue]:
        """Returns an iterator for values of the column at the tiven index."""
        return (row[column_index] for row in self.rows)

    def _sorted_by_column_name(self) -> MetricFlowDataTable:  # noqa: D102
        new_rows: List[List[CellValue]] = [[] for _ in range(self.row_count)]
        sorted_column_names = sorted(self.column_names)
        for column_name in sorted_column_names:
            old_column_index = self.column_name_index(column_name)
            for row_index, cell_value in enumerate(self.column_values_iterator(old_column_index)):
                new_rows[row_index].append(cell_value)

        return MetricFlowDataTable(
            column_descriptions=tuple(
                self.column_descriptions[self.column_name_index(column_name)] for column_name in sorted_column_names
            ),
            rows=tuple(
                tuple(row_dict[column_index] for column_index in range(self.column_count)) for row_dict in new_rows
            ),
        )

    def _sorted_by_row(self) -> MetricFlowDataTable:  # noqa: D102
        def _cell_sort_key(cell: CellValue) -> str:
            if isinstance(cell, datetime.datetime):
                return cell.isoformat()
            return str(cell)

        def _row_sort_key(row: Tuple[CellValue, ...]) -> Tuple[str, ...]:
            return tuple(_cell_sort_key(cell) for cell in row)

        return MetricFlowDataTable(
            column_descriptions=self.column_descriptions,
            rows=tuple(sorted((row for row in self.rows), key=_row_sort_key)),
        )

    def sorted(self) -> MetricFlowDataTable:
        """Returns this but with the columns in order by name, and the rows in order by values."""
        return self._sorted_by_column_name()._sorted_by_row()

    def text_format(self, float_decimals: int = 2) -> str:
        """Return a text version of this table that is suitable for printing."""
        str_rows: List[List[str]] = []
        for row in self.rows:
            str_row: List[str] = []
            for cell_value in row:
                if isinstance(cell_value, float):
                    str_row.append(f"{cell_value:.{float_decimals}f}")
                    continue

                if isinstance(cell_value, datetime.datetime):
                    str_row.append(cell_value.isoformat())
                    continue

                str_row.append(str(cell_value))
            str_rows.append(str_row)
        return tabulate.tabulate(
            tabular_data=tuple(tuple(str_row) for str_row in str_rows),
            headers=tuple(column_description.column_name for column_description in self.column_descriptions),
        )

    def with_lower_case_column_names(self) -> MetricFlowDataTable:
        """Return this but with columns names in lowercase."""
        return MetricFlowDataTable(
            column_descriptions=tuple(
                column_description.with_lower_case_column_name() for column_description in self.column_descriptions
            ),
            rows=self.rows,
        )

    def get_cell_value(self, row_index: int, column_index: int) -> CellValue:  # noqa: D102
        return self.rows[row_index][column_index]

    @staticmethod
    def create_from_rows(  # noqa: D102
        column_names: Sequence[str], rows: Iterable[Sequence[InputCellValue]]
    ) -> MetricFlowDataTable:
        builder = _MetricFlowDataTableBuilder(column_names)
        for row in rows:
            builder.add_row(row)
        return builder.build()


class _MetricFlowDataTableBuilder:
    """Helps build `MetricFlowDataTable`, one row at a time.

    This validates each row as it is input to give better error messages.
    """

    def __init__(self, column_names: Sequence[str]) -> None:  # noqa: D107
        self._rows: List[Tuple[CellValue, ...]] = []
        self._column_names = tuple(column_names)

    def _build_table_from_rows(self) -> MetricFlowDataTable:  # noqa: D102
        # Figure out the type of the column based on the types of the values in the rows.
        # Can't use the type of the columns in the first row because it might contain None, so iterate through the rows
        # and use the first non-None type.
        column_types_so_far: Optional[Tuple[Type[CellValue], ...]] = None
        cell_value: CellValue
        for row in self._rows:
            if column_types_so_far is None:
                column_types_so_far = row_cell_types(row)
                continue

            # If the types of the objects in the row are the same, no need for updates.
            row_column_types = row_cell_types(row)
            if row_column_types == column_types_so_far:
                continue

            # Types of objects in the row are different from what's known so far.
            # They can only be different in that one can be None and other can be not None.
            updated_column_types: List[Type[CellValue]] = []
            for column_type_so_far, cell_type, cell_value in itertools.zip_longest(
                column_types_so_far, row_column_types, row
            ):
                if column_type_so_far is cell_type:
                    updated_column_types.append(column_type_so_far)
                elif column_type_so_far is type(None) and cell_type is not None:
                    updated_column_types.append(cell_type)
                elif column_type_so_far is not None and cell_type is type(None):
                    updated_column_types.append(column_type_so_far)
                else:
                    raise ValueError(f"Expected cell type {column_type_so_far} but got: {cell_type}")

            column_types_so_far = tuple(updated_column_types)

        # Empty table case.
        if column_types_so_far is None:
            column_types_so_far = tuple(type(None) for _ in range(len(self._column_names)))

        final_column_types = column_types_so_far

        return MetricFlowDataTable(
            column_descriptions=tuple(
                ColumnDescription(column_name=column_name, column_type=column_type)
                for column_name, column_type in itertools.zip_longest(self._column_names, final_column_types)
            ),
            rows=tuple(self._rows),
        )

    def _convert_row_to_supported_types(self, row: Sequence[InputCellValue]) -> Sequence[CellValue]:
        """Since only a limited set of types are supported, convert the input type to the supported type."""
        updated_row: List[CellValue] = []
        for cell_value in row:
            if (
                cell_value is None
                or isinstance(cell_value, float)
                or isinstance(cell_value, bool)
                or isinstance(cell_value, int)
                or isinstance(cell_value, str)
            ):
                updated_row.append(cell_value)
                continue

            if isinstance(cell_value, datetime.datetime):
                updated_row.append(cell_value.replace(tzinfo=None))
                continue

            if isinstance(cell_value, Decimal):
                updated_row.append(float(cell_value))
                continue

            if isinstance(cell_value, datetime.date):
                updated_row.append(datetime.datetime.combine(cell_value, datetime.datetime.min.time()))
                continue

            raise ValueError(f"Row cell has unexpected type: {repr(cell_value)}")
        return updated_row

    def add_row(self, row: Sequence[InputCellValue], parse_strings: bool = False) -> Self:  # noqa: D102
        row = tuple(row)
        expected_column_count = len(self._column_names)
        actual_column_count = len(row)
        if actual_column_count != expected_column_count:
            raise ValueError(
                f"Input row has {actual_column_count} columns, but expected {expected_column_count} columns. Row is:"
                f"\n{indent(mf_pformat(row))}"
            )
        self._rows.append(tuple(self._convert_row_to_supported_types(row)))
        return self

    def build(self) -> MetricFlowDataTable:  # noqa: D102
        return self._build_table_from_rows()



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

from typing_extensions import override

from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryIssueType,
    MetricFlowQueryResolutionIssue,
)
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput


@dataclass(frozen=True)
class NoMatchingItemsForNoMetricsQuery(MetricFlowQueryResolutionIssue):
    """Describes an issue with the query where there are no matching items for a no-metrics / distinct values query."""

    @staticmethod
    def from_parameters(  # noqa: D102
        parent_issues: Sequence[MetricFlowQueryResolutionIssue],
        query_resolution_path: MetricFlowQueryResolutionPath,
    ) -> NoMatchingItemsForNoMetricsQuery:
        return NoMatchingItemsForNoMetricsQuery(
            issue_type=MetricFlowQueryIssueType.ERROR,
            parent_issues=tuple(parent_issues),
            query_resolution_path=query_resolution_path,
        )

    @override
    def ui_description(self, associated_input: MetricFlowQueryResolverInput) -> str:
        # Could be useful to list the candidates, but it makes the logs very long.
        return (
            "The given input does not match any of the available group-by-items for a distinct\n"
            "values query without metrics."
        )

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> NoMatchingItemsForNoMetricsQuery:
        return NoMatchingItemsForNoMetricsQuery(
            issue_type=self.issue_type,
            parent_issues=tuple(issue.with_path_prefix(path_prefix) for issue in self.parent_issues),
            query_resolution_path=self.query_resolution_path.with_path_prefix(path_prefix),
        )



from __future__ import annotations

from typing import List, Sequence

from metricflow_semantics.specs.column_assoc import ColumnAssociation, ColumnAssociationResolver
from metricflow_semantics.specs.spec_set import InstanceSpecSet, InstanceSpecSetTransform

from metricflow.plan_conversion.select_column_gen import SelectColumnSet
from metricflow.plan_conversion.sql_expression_builders import make_coalesced_expr
from metricflow.sql.sql_plan import SqlSelectColumn


class CreateSelectCoalescedColumnsForLinkableSpecs(InstanceSpecSetTransform[SelectColumnSet]):
    """Create select columns that coalesce columns corresponding to linkable specs.

    e.g.

    dimension_specs = [DimensionSpec(element_name="is_instant")]
    table_aliases = ["a", "b"]

    ->

    COALESCE(a.is_instant, b.is_instant) AS is_instant
    """

    def __init__(  # noqa: D107
        self,
        column_association_resolver: ColumnAssociationResolver,
        table_aliases: Sequence[str],
    ) -> None:
        self._column_association_resolver = column_association_resolver
        self._table_aliases = table_aliases

    def transform(self, spec_set: InstanceSpecSet) -> SelectColumnSet:  # noqa: D102
        dimension_columns: List[SqlSelectColumn] = []
        time_dimension_columns: List[SqlSelectColumn] = []
        entity_columns: List[SqlSelectColumn] = []

        for dimension_spec in spec_set.dimension_specs:
            column_name = self._column_association_resolver.resolve_spec(dimension_spec).column_name
            dimension_columns.append(
                SqlSelectColumn(
                    expr=make_coalesced_expr(self._table_aliases, column_name),
                    column_alias=column_name,
                )
            )

        for time_dimension_spec in spec_set.time_dimension_specs:
            column_name = self._column_association_resolver.resolve_spec(time_dimension_spec).column_name
            time_dimension_columns.append(
                SqlSelectColumn(
                    expr=make_coalesced_expr(self._table_aliases, column_name),
                    column_alias=column_name,
                )
            )

        for entity_spec in spec_set.entity_specs:
            column_name = self._column_association_resolver.resolve_spec(entity_spec).column_name

            entity_columns.append(
                SqlSelectColumn(
                    expr=make_coalesced_expr(self._table_aliases, column_name),
                    column_alias=column_name,
                )
            )

        return SelectColumnSet(
            dimension_columns=dimension_columns,
            time_dimension_columns=time_dimension_columns,
            entity_columns=entity_columns,
        )


class SelectOnlyLinkableSpecs(InstanceSpecSetTransform[InstanceSpecSet]):
    """Removes metrics and measures from the spec set."""

    def transform(self, spec_set: InstanceSpecSet) -> InstanceSpecSet:  # noqa: D102
        return InstanceSpecSet(
            metric_specs=(),
            measure_specs=(),
            dimension_specs=spec_set.dimension_specs,
            time_dimension_specs=spec_set.time_dimension_specs,
            entity_specs=spec_set.entity_specs,
        )


class CreateColumnAssociations(InstanceSpecSetTransform[Sequence[ColumnAssociation]]):
    """Using the specs in the instance set, generate a list of the associated column associations.

    Initial use case is to figure out names of the columns present in the SQL of a WhereFilter.
    """

    def __init__(self, column_association_resolver: ColumnAssociationResolver) -> None:  # noqa: D107
        self._column_association_resolver = column_association_resolver

    def transform(self, spec_set: InstanceSpecSet) -> Sequence[ColumnAssociation]:  # noqa: D102
        return tuple(self._column_association_resolver.resolve_spec(spec) for spec in spec_set.all_specs)



from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, List, Sequence, Tuple, TypeVar

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass
from dbt_semantic_interfaces.references import EntityReference, LinkableElementReference

from metricflow_semantics.model.semantics.linkable_element import ElementPathKey
from metricflow_semantics.naming.linkable_spec_name import StructuredLinkableSpecName

if typing.TYPE_CHECKING:
    from metricflow_semantics.specs.dimension_spec import DimensionSpec
    from metricflow_semantics.specs.entity_spec import EntitySpec
    from metricflow_semantics.specs.group_by_metric_spec import GroupByMetricSpec
    from metricflow_semantics.specs.measure_spec import MeasureSpec
    from metricflow_semantics.specs.metadata_spec import MetadataSpec
    from metricflow_semantics.specs.metric_spec import MetricSpec
    from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec
from metricflow_semantics.visitor import VisitorOutputT


@dataclass(frozen=True)
class InstanceSpec(SerializableDataclass):
    """A specification for an instance of a metric definition object.

    An instance is different from the definition object in that it correlates to columns in the data flow and can be in
    different states. e.g. a time dimension at a different time granularity.

    This can't be a Protocol as base classes of Protocols need to be Protocols.
    """

    """Name of the dimension or entity in the semantic model."""
    element_name: str

    @staticmethod
    def merge(*specs: Sequence[InstanceSpec]) -> List[InstanceSpec]:
        """Merge all specs into a single list."""
        result: List[InstanceSpec] = []
        for spec in specs:
            result.extend(spec)
        return result

    @property
    def qualified_name(self) -> str:
        """Return the qualified name of this spec. e.g. "user_id__country"."""
        raise NotImplementedError()

    def accept(self, visitor: InstanceSpecVisitor[VisitorOutputT]) -> VisitorOutputT:
        """See Visitable."""
        raise NotImplementedError()

    def without_filter_specs(self) -> InstanceSpec:
        """Return the instance spec without any filtering (for comparison purposes)."""
        return self


class InstanceSpecVisitor(Generic[VisitorOutputT], ABC):
    """Visitor for the InstanceSpec classes."""

    @abstractmethod
    def visit_measure_spec(self, measure_spec: MeasureSpec) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_dimension_spec(self, dimension_spec: DimensionSpec) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_time_dimension_spec(self, time_dimension_spec: TimeDimensionSpec) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_entity_spec(self, entity_spec: EntitySpec) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_group_by_metric_spec(self, group_by_metric_spec: GroupByMetricSpec) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_metric_spec(self, metric_spec: MetricSpec) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_metadata_spec(self, metadata_spec: MetadataSpec) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError


@dataclass(frozen=True)
class LinkableInstanceSpec(InstanceSpec, ABC):
    """Generally a dimension or entity that may be specified using entity links.

    For example, user_id__country -> LinkableElementSpec(element_name="country", entity_links=["user_id"]

    See InstanceSpec for the reason behind "type: ignore"
    """

    """A list representing the join path of entities to get to this element."""
    entity_links: Tuple[EntityReference, ...]

    @property
    @abstractmethod
    def without_first_entity_link(self: SelfTypeT) -> SelfTypeT:
        """e.g. user_id__device_id__platform -> device_id__platform."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def without_entity_links(self: SelfTypeT) -> SelfTypeT:
        """e.g. user_id__device_id__platform -> platform."""
        raise NotImplementedError()

    @staticmethod
    def merge_linkable_specs(*specs: Sequence[LinkableInstanceSpec]) -> List[LinkableInstanceSpec]:
        """Merge all specs into a single list."""
        result: List[LinkableInstanceSpec] = []
        for spec in specs:
            result.extend(spec)
        return result

    @property
    def qualified_name(self) -> str:
        """Return the qualified name of this spec. e.g. "user_id__country"."""
        return StructuredLinkableSpecName(
            entity_link_names=tuple(x.element_name for x in self.entity_links), element_name=self.element_name
        ).qualified_name

    @property
    @abstractmethod
    def reference(self) -> LinkableElementReference:
        """Return the LinkableElementReference associated with the spec instance."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def element_path_key(self) -> ElementPathKey:
        """Return the ElementPathKey representation of the LinkableInstanceSpec subtype."""
        raise NotImplementedError()

    @abstractmethod
    def with_entity_prefix(self, entity_prefix: EntityReference) -> LinkableInstanceSpec:
        """Add the selected entity prefix to the start of the entity links."""
        raise NotImplementedError()


SelfTypeT = TypeVar("SelfTypeT", bound="LinkableInstanceSpec")



from __future__ import annotations

from typing import Optional

from dbt_semantic_interfaces.references import MetricReference
from typing_extensions import override

from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.naming.naming_scheme import QueryItemNamingScheme
from metricflow_semantics.specs.instance_spec import InstanceSpec
from metricflow_semantics.specs.patterns.metric_pattern import MetricSpecPattern
from metricflow_semantics.specs.spec_set import group_spec_by_type


class MetricNamingScheme(QueryItemNamingScheme):
    """A naming scheme for metrics."""

    @override
    def input_str(self, instance_spec: InstanceSpec) -> Optional[str]:
        spec_set = group_spec_by_type(instance_spec)
        names = tuple(spec.element_name for spec in spec_set.metric_specs)

        if len(names) != 1:
            raise RuntimeError(f"Did not get 1 name for {instance_spec}. Got {names}")

        return names[0]

    @override
    def spec_pattern(self, input_str: str, semantic_manifest_lookup: SemanticManifestLookup) -> MetricSpecPattern:
        input_str = input_str.lower()
        if not self.input_str_follows_scheme(input_str):
            raise RuntimeError(f"{repr(input_str)} does not follow this scheme.")
        return MetricSpecPattern(metric_reference=MetricReference(element_name=input_str))

    @override
    def input_str_follows_scheme(self, input_str: str) -> bool:
        # TODO: Use regex.
        return True

    @override
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id()={hex(id(self))})"



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Sequence

from dbt_semantic_interfaces.references import TimeDimensionReference
from dbt_semantic_interfaces.type_enums import TimeGranularity
from metricflow_semantics.specs.measure_spec import MeasureSpec
from metricflow_semantics.specs.non_additive_dimension_spec import NonAdditiveDimensionSpec


@dataclass(frozen=True)
class MeasureSpecProperties:
    """Input dataclass for grouping properties of a sequence of MeasureSpecs."""

    measure_specs: Sequence[MeasureSpec]
    semantic_model_name: str
    agg_time_dimension: TimeDimensionReference
    agg_time_dimension_grain: TimeGranularity
    non_additive_dimension_spec: Optional[NonAdditiveDimensionSpec] = None



from __future__ import annotations

import itertools
import logging
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, List, Optional, Sequence, Set, Tuple

from dbt_semantic_interfaces.references import MeasureReference, MetricReference, SemanticModelReference

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.mf_logging.pretty_print import mf_pformat, mf_pformat_dict
from metricflow_semantics.mf_logging.runtime import log_runtime
from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.model.semantic_model_derivation import SemanticModelDerivation
from metricflow_semantics.model.semantics.linkable_element_set import LinkableElementSet
from metricflow_semantics.naming.metric_scheme import MetricNamingScheme
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_pattern_factory import (
    WhereFilterPatternFactory,
)
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_spec_lookup import (
    FilterSpecResolutionLookUp,
)
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_spec_resolver import (
    WhereFilterSpecResolver,
)
from metricflow_semantics.query.group_by_item.group_by_item_resolver import GroupByItemResolution, GroupByItemResolver
from metricflow_semantics.query.group_by_item.resolution_dag.dag import GroupByItemResolutionDag
from metricflow_semantics.query.group_by_item.resolution_dag.dag_builder import GroupByItemResolutionDagBuilder
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.query_resolution_node import (
    QueryGroupByItemResolutionNode,
)
from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryResolutionIssueSet,
)
from metricflow_semantics.query.issues.parsing.duplicate_metric_alias import DuplicateMetricAliasIssue
from metricflow_semantics.query.issues.parsing.invalid_limit import InvalidLimitIssue
from metricflow_semantics.query.issues.parsing.invalid_metric import InvalidMetricIssue
from metricflow_semantics.query.issues.parsing.invalid_min_max_only import InvalidMinMaxOnlyIssue
from metricflow_semantics.query.issues.parsing.invalid_order import InvalidOrderByItemIssue
from metricflow_semantics.query.issues.parsing.no_metric_or_group_by import NoMetricOrGroupByIssue
from metricflow_semantics.query.query_resolution import (
    InputToIssueSetMapping,
    InputToIssueSetMappingItem,
    MetricFlowQueryResolution,
)
from metricflow_semantics.query.resolver_inputs.query_resolver_inputs import (
    ResolverInputForGroupByItem,
    ResolverInputForLimit,
    ResolverInputForMetric,
    ResolverInputForMinMaxOnly,
    ResolverInputForOrderByItem,
    ResolverInputForQuery,
    ResolverInputForQueryLevelWhereFilterIntersection,
    ResolverInputForWhereFilterIntersection,
)
from metricflow_semantics.query.suggestion_generator import QueryItemSuggestionGenerator
from metricflow_semantics.query.validation_rules.duplicate_metric import DuplicateMetricValidationRule
from metricflow_semantics.query.validation_rules.metric_time_requirements import MetricTimeQueryValidationRule
from metricflow_semantics.query.validation_rules.query_validator import PostResolutionQueryValidator
from metricflow_semantics.specs.instance_spec import InstanceSpec, LinkableInstanceSpec
from metricflow_semantics.specs.metric_spec import MetricSpec
from metricflow_semantics.specs.order_by_spec import OrderBySpec
from metricflow_semantics.specs.query_spec import MetricFlowQuerySpec
from metricflow_semantics.specs.spec_set import group_specs_by_type
from metricflow_semantics.workarounds.reference import sorted_semantic_model_references

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ResolveOrderByResult:
    """Result of resolving order-by inputs to specs."""

    order_by_specs: Tuple[OrderBySpec, ...]
    input_to_issue_set_mapping: InputToIssueSetMapping


@dataclass(frozen=True)
class ResolveMetricsResult:
    """Result of resolving a metric input to specs."""

    metric_specs: Tuple[MetricSpec, ...]
    input_to_issue_set_mapping: InputToIssueSetMapping


@dataclass(frozen=True)
class ResolveLimitResult:
    """Result of resolving a limit input."""

    limit: Optional[int]
    input_to_issue_set_mapping: InputToIssueSetMapping


@dataclass(frozen=True)
class ResolveMinMaxOnlyResult:
    """Result of resolving a limit input."""

    min_max_only: bool
    input_to_issue_set_mapping: InputToIssueSetMapping


@dataclass(frozen=True)
class ResolveGroupByItemsResult:
    """Result of resolving group-by inputs to specs."""

    resolution_dag: GroupByItemResolutionDag
    group_by_item_specs: Tuple[LinkableInstanceSpec, ...]
    input_to_issue_set_mapping: InputToIssueSetMapping
    linkable_element_set: LinkableElementSet


@dataclass(frozen=True)
class ResolveMetricOrGroupByItemsResult:
    """Result of checking that there are metrics or group by items in the query."""

    input_to_issue_set_mapping: InputToIssueSetMapping


class MetricFlowQueryResolver:
    """Resolves inputs to a query (e.g. metrics, group by items into concrete specs."""

    def __init__(  # noqa: D107
        self,
        manifest_lookup: SemanticManifestLookup,
        where_filter_pattern_factory: WhereFilterPatternFactory,
    ) -> None:
        self._manifest_lookup = manifest_lookup
        self._post_resolution_query_validator = PostResolutionQueryValidator()
        self._where_filter_pattern_factory = where_filter_pattern_factory

    @staticmethod
    def _resolve_has_metric_or_group_by_inputs(
        resolver_input_for_query: ResolverInputForQuery, query_resolution_path: MetricFlowQueryResolutionPath
    ) -> ResolveMetricOrGroupByItemsResult:
        if len(resolver_input_for_query.metric_inputs) == 0 and len(resolver_input_for_query.group_by_item_inputs) == 0:
            return ResolveMetricOrGroupByItemsResult(
                input_to_issue_set_mapping=InputToIssueSetMapping.from_one_item(
                    resolver_input=resolver_input_for_query,
                    issue_set=MetricFlowQueryResolutionIssueSet.from_issue(
                        NoMetricOrGroupByIssue.from_parameters(
                            resolver_input_for_query=resolver_input_for_query,
                            query_resolution_path=query_resolution_path,
                        ),
                    ),
                )
            )
        return ResolveMetricOrGroupByItemsResult(input_to_issue_set_mapping=InputToIssueSetMapping.empty_instance())

    @staticmethod
    def _resolve_group_by_item_input(
        group_by_item_input: ResolverInputForGroupByItem,
        group_by_item_resolver: GroupByItemResolver,
    ) -> GroupByItemResolution:
        suggestion_generator = QueryItemSuggestionGenerator(
            input_naming_scheme=group_by_item_input.input_obj_naming_scheme,
            input_str=str(group_by_item_input.input_obj),
            candidate_filters=QueryItemSuggestionGenerator.GROUP_BY_ITEM_CANDIDATE_FILTERS,
        )
        return group_by_item_resolver.resolve_matching_item_for_querying(
            spec_pattern=group_by_item_input.spec_pattern,
            suggestion_generator=suggestion_generator,
        )

    def _resolve_metric_inputs(
        self,
        metric_inputs: Sequence[ResolverInputForMetric],
        query_resolution_path: MetricFlowQueryResolutionPath,
    ) -> ResolveMetricsResult:
        # Build a list of metrics that are available from the manifest.
        available_metric_specs = tuple(
            MetricSpec.from_reference(metric_reference)
            for metric_reference in self._manifest_lookup.metric_lookup.metric_references
        )
        metric_specs: List[MetricSpec] = []
        input_to_issue_set_mapping_items: List[InputToIssueSetMappingItem] = []
        alias_to_metrics: Dict[str, List[Tuple[ResolverInputForMetric, MetricReference]]] = defaultdict(list)

        # Find the metric that matches the metric pattern from the input.
        for metric_input in metric_inputs:
            matching_specs = metric_input.spec_pattern.match(available_metric_specs)
            if len(matching_specs) == 1:
                matching_spec = matching_specs[0]
                alias = metric_input.alias
                if alias:
                    matching_spec = matching_spec.with_alias(alias)
                metric_specs.append(matching_spec)
                resolved_name = matching_spec.alias or matching_spec.qualified_name
                alias_to_metrics[resolved_name].append((metric_input, matching_spec.reference))
            else:
                suggestion_generator = QueryItemSuggestionGenerator(
                    input_naming_scheme=MetricNamingScheme(),
                    input_str=str(metric_input.input_obj),
                    candidate_filters=(),
                )
                metric_suggestions = suggestion_generator.input_suggestions(candidate_specs=available_metric_specs)
                input_to_issue_set_mapping_items.append(
                    InputToIssueSetMappingItem(
                        resolver_input=metric_input,
                        issue_set=MetricFlowQueryResolutionIssueSet.from_issue(
                            InvalidMetricIssue.from_parameters(
                                metric_suggestions=metric_suggestions,
                                query_resolution_path=query_resolution_path,
                            )
                        ),
                    )
                )

        # Find any duplicate aliases
        for alias, metrics in alias_to_metrics.items():
            if len(metrics) > 1:
                metric_inputs = [m[0] for m in metrics]
                metric_references = [m[1] for m in metrics]
                input_to_issue_set_mapping_items.append(
                    InputToIssueSetMappingItem(
                        resolver_input=metric_inputs[0],
                        issue_set=MetricFlowQueryResolutionIssueSet.from_issue(
                            DuplicateMetricAliasIssue.from_parameters(
                                duplicate_metric_references=metric_references,
                                query_resolution_path=query_resolution_path,
                            )
                        ),
                    )
                )

        return ResolveMetricsResult(
            metric_specs=tuple(metric_specs),
            input_to_issue_set_mapping=InputToIssueSetMapping(items=tuple(input_to_issue_set_mapping_items)),
        )

    def _resolve_group_by_items_result(
        self,
        metric_references: Sequence[MetricReference],
        group_by_item_inputs: Sequence[ResolverInputForGroupByItem],
        filter_input: ResolverInputForQueryLevelWhereFilterIntersection,
    ) -> ResolveGroupByItemsResult:
        resolution_dag_builder = GroupByItemResolutionDagBuilder(
            manifest_lookup=self._manifest_lookup,
        )
        resolution_dag = resolution_dag_builder.build(
            metric_references=metric_references,
            where_filter_intersection=filter_input.where_filter_intersection,
        )
        logger.debug(LazyFormat(lambda: f"Resolution DAG is:\n{resolution_dag.structure_text()}"))

        group_by_item_resolver = GroupByItemResolver(
            manifest_lookup=self._manifest_lookup,
            resolution_dag=resolution_dag,
        )

        input_to_issue_set_mapping_items: List[InputToIssueSetMappingItem] = []
        group_by_item_specs: List[LinkableInstanceSpec] = []
        linkable_element_sets: List[LinkableElementSet] = []
        for group_by_item_input in group_by_item_inputs:
            resolution = MetricFlowQueryResolver._resolve_group_by_item_input(
                group_by_item_resolver=group_by_item_resolver,
                group_by_item_input=group_by_item_input,
            )
            if resolution.issue_set.has_issues:
                input_to_issue_set_mapping_items.append(
                    InputToIssueSetMappingItem(resolver_input=group_by_item_input, issue_set=resolution.issue_set)
                )
            if resolution.spec is not None:
                group_by_item_specs.append(resolution.spec)
                linkable_element_sets.append(resolution.linkable_element_set)

        return ResolveGroupByItemsResult(
            resolution_dag=resolution_dag,
            group_by_item_specs=tuple(group_by_item_specs),
            input_to_issue_set_mapping=InputToIssueSetMapping(tuple(input_to_issue_set_mapping_items)),
            linkable_element_set=LinkableElementSet.merge_by_path_key(linkable_element_sets),
        )

    @staticmethod
    def _resolve_order_by(
        resolver_inputs_for_order_by_items: Sequence[ResolverInputForOrderByItem],
        metric_specs: Sequence[MetricSpec],
        group_by_item_specs: Sequence[LinkableInstanceSpec],
        query_resolution_path: MetricFlowQueryResolutionPath,
    ) -> ResolveOrderByResult:
        mapping_items: List[InputToIssueSetMappingItem] = []
        order_by_specs: List[OrderBySpec] = []

        # Match the pattern from the order by input to one of the metric or group-by-item specs.
        # The pattern needs to be used because there are cases where the order-by-item is specified in a different way
        # from the group-by-item, so an equality comparison won't work.
        for resolver_input_for_order_by in resolver_inputs_for_order_by_items:
            matching_specs: List[InstanceSpec] = []
            for possible_input in resolver_input_for_order_by.possible_inputs:
                spec_pattern = possible_input.spec_pattern
                matching_specs.extend(spec_pattern.match(metric_specs))
                matching_specs.extend(spec_pattern.match(group_by_item_specs))

            if len(matching_specs) != 1:
                mapping_items.append(
                    InputToIssueSetMappingItem(
                        resolver_input=resolver_input_for_order_by,
                        issue_set=MetricFlowQueryResolutionIssueSet.from_issue(
                            InvalidOrderByItemIssue.from_parameters(
                                order_by_item_input=resolver_input_for_order_by,
                                query_resolution_path=query_resolution_path,
                            )
                        ),
                    )
                )
            else:
                order_by_specs.append(
                    OrderBySpec(
                        instance_spec=matching_specs[0],
                        descending=resolver_input_for_order_by.descending,
                    )
                )

        return ResolveOrderByResult(
            input_to_issue_set_mapping=InputToIssueSetMapping(items=tuple(mapping_items)),
            order_by_specs=tuple(order_by_specs),
        )

    @staticmethod
    def _resolve_limit_input(
        limit_input: ResolverInputForLimit, query_resolution_path: MetricFlowQueryResolutionPath
    ) -> ResolveLimitResult:
        limit = limit_input.limit
        if limit is not None and limit < 0:
            return ResolveLimitResult(
                limit=limit,
                input_to_issue_set_mapping=InputToIssueSetMapping.from_one_item(
                    resolver_input=limit_input,
                    issue_set=MetricFlowQueryResolutionIssueSet.from_issue(
                        InvalidLimitIssue.from_parameters(limit=limit, query_resolution_path=query_resolution_path),
                    ),
                ),
            )
        return ResolveLimitResult(limit=limit, input_to_issue_set_mapping=InputToIssueSetMapping.empty_instance())

    @staticmethod
    def _resolve_min_max_only_input(
        min_max_only_input: ResolverInputForMinMaxOnly,
        query_resolution_path: MetricFlowQueryResolutionPath,
        metric_inputs: Tuple[ResolverInputForMetric, ...],
        group_by_item_inputs: Tuple[ResolverInputForGroupByItem, ...],
        order_by_item_inputs: Tuple[ResolverInputForOrderByItem, ...],
        limit_input: ResolverInputForLimit,
    ) -> ResolveMinMaxOnlyResult:
        min_max_only = min_max_only_input.min_max_only
        if min_max_only:
            if (
                metric_inputs
                or order_by_item_inputs
                or (limit_input.limit is not None)
                or len(group_by_item_inputs) != 1
            ):
                return ResolveMinMaxOnlyResult(
                    min_max_only=min_max_only,
                    input_to_issue_set_mapping=InputToIssueSetMapping.from_one_item(
                        resolver_input=min_max_only_input,
                        issue_set=MetricFlowQueryResolutionIssueSet.from_issue(
                            InvalidMinMaxOnlyIssue.from_parameters(
                                min_max_only=min_max_only, query_resolution_path=query_resolution_path
                            ),
                        ),
                    ),
                )
        return ResolveMinMaxOnlyResult(
            min_max_only=min_max_only, input_to_issue_set_mapping=InputToIssueSetMapping.empty_instance()
        )

    def _build_filter_spec_lookup(
        self,
        resolution_dag: GroupByItemResolutionDag,
    ) -> FilterSpecResolutionLookUp:
        where_filter_spec_resolver = WhereFilterSpecResolver(
            manifest_lookup=self._manifest_lookup,
            resolution_dag=resolution_dag,
            spec_pattern_factory=self._where_filter_pattern_factory,
        )

        return where_filter_spec_resolver.resolve_lookup()

    def resolve_query(self, resolver_input_for_query: ResolverInputForQuery) -> MetricFlowQueryResolution:
        """Resolve the query into specs that can be passed into the next stage in query processing."""
        # Workaround for a Pycharm type inspection issue with decorators.
        # noinspection PyArgumentList
        return self._resolve_query(resolver_input_for_query=resolver_input_for_query)

    @log_runtime()
    def _resolve_query(self, resolver_input_for_query: ResolverInputForQuery) -> MetricFlowQueryResolution:
        metric_inputs = resolver_input_for_query.metric_inputs
        group_by_item_inputs = resolver_input_for_query.group_by_item_inputs
        order_by_item_inputs = resolver_input_for_query.order_by_item_inputs
        limit_input = resolver_input_for_query.limit_input
        query_level_filter_input = resolver_input_for_query.filter_input
        min_max_only_input = resolver_input_for_query.min_max_only

        # Define a resolution path for issues where the input is considered to be the whole query.
        query_resolution_path = MetricFlowQueryResolutionPath.from_path_item(
            QueryGroupByItemResolutionNode.create(
                parent_nodes=(),
                metrics_in_query=tuple(metric_input.spec_pattern.metric_reference for metric_input in metric_inputs),
                where_filter_intersection=query_level_filter_input.where_filter_intersection,
            )
        )

        mappings_to_merge: List[InputToIssueSetMapping] = []

        # Check that the query contains metrics or group by items.
        resolve_metrics_or_group_by_result = self._resolve_has_metric_or_group_by_inputs(
            resolver_input_for_query=resolver_input_for_query, query_resolution_path=query_resolution_path
        )
        mappings_to_merge.append(resolve_metrics_or_group_by_result.input_to_issue_set_mapping)

        # Resolve metrics.
        resolve_metrics_result = self._resolve_metric_inputs(
            metric_inputs=metric_inputs,
            query_resolution_path=query_resolution_path,
        )
        mappings_to_merge.append(resolve_metrics_result.input_to_issue_set_mapping)
        metric_specs = resolve_metrics_result.metric_specs

        # Resolve limit
        resolve_limit_result = self._resolve_limit_input(
            limit_input=limit_input,
            query_resolution_path=query_resolution_path,
        )
        mappings_to_merge.append(resolve_limit_result.input_to_issue_set_mapping)

        # Resolve min max only
        resolve_min_max_only_result = self._resolve_min_max_only_input(
            min_max_only_input=min_max_only_input,
            query_resolution_path=query_resolution_path,
            metric_inputs=metric_inputs,
            group_by_item_inputs=group_by_item_inputs,
            order_by_item_inputs=order_by_item_inputs,
            limit_input=limit_input,
        )
        mappings_to_merge.append(resolve_min_max_only_result.input_to_issue_set_mapping)

        # Early stop before resolving further as with invalid metrics, the errors won't be as useful.
        issue_set_mapping_so_far = InputToIssueSetMapping.merge_iterable(mappings_to_merge)
        if issue_set_mapping_so_far.has_issues:
            return MetricFlowQueryResolution(
                query_spec=None,
                resolution_dag=None,
                filter_spec_lookup=FilterSpecResolutionLookUp.empty_instance(),
                input_to_issue_set=issue_set_mapping_so_far,
                queried_semantic_models=(),
            )

        # Resolve group by items.
        resolve_group_by_item_result = self._resolve_group_by_items_result(
            metric_references=tuple(metric_spec.reference for metric_spec in metric_specs),
            group_by_item_inputs=group_by_item_inputs,
            filter_input=query_level_filter_input,
        )
        resolution_dag = resolve_group_by_item_result.resolution_dag
        group_by_item_specs = resolve_group_by_item_result.group_by_item_specs
        mappings_to_merge.append(resolve_group_by_item_result.input_to_issue_set_mapping)

        # Resolve order by.
        resolve_order_by_result = MetricFlowQueryResolver._resolve_order_by(
            resolver_inputs_for_order_by_items=order_by_item_inputs,
            metric_specs=tuple(metric_spec.with_alias(None) for metric_spec in metric_specs),
            group_by_item_specs=group_by_item_specs,
            query_resolution_path=query_resolution_path,
        )
        order_by_specs = resolve_order_by_result.order_by_specs
        if resolve_order_by_result.input_to_issue_set_mapping.has_issues:
            mappings_to_merge.append(resolve_order_by_result.input_to_issue_set_mapping)

        # Resolve all where filters in the DAG and generate mappings if there are issues.
        filter_spec_lookup = self._build_filter_spec_lookup(resolution_dag)
        for filter_spec_resolution in filter_spec_lookup.spec_resolutions:
            if filter_spec_resolution.issue_set.has_issues:
                mappings_to_merge.append(
                    InputToIssueSetMapping.from_one_item(
                        resolver_input=ResolverInputForWhereFilterIntersection(
                            filter_resolution_path=filter_spec_resolution.filter_location_path,
                            where_filter_intersection=filter_spec_resolution.where_filter_intersection,
                            object_builder_str=filter_spec_resolution.object_builder_str,
                        ),
                        issue_set=filter_spec_resolution.issue_set,
                    )
                )

        for non_parseable_resolution in filter_spec_lookup.non_parsable_resolutions:
            filter_resolver_input = ResolverInputForWhereFilterIntersection(
                filter_resolution_path=non_parseable_resolution.filter_location_path,
                where_filter_intersection=non_parseable_resolution.where_filter_intersection,
                object_builder_str=None,
            )
            mappings_to_merge.append(
                InputToIssueSetMapping.from_one_item(
                    resolver_input=filter_resolver_input,
                    issue_set=non_parseable_resolution.issue_set,
                )
            )

        # Return if there are any errors as issues generated by later validations may not make sense otherwise.
        issue_set_mapping_so_far = InputToIssueSetMapping.merge_iterable(mappings_to_merge)
        if issue_set_mapping_so_far.has_issues:
            return MetricFlowQueryResolution(
                query_spec=None,
                resolution_dag=resolution_dag,
                filter_spec_lookup=filter_spec_lookup,
                input_to_issue_set=issue_set_mapping_so_far,
                queried_semantic_models=(),
            )

        # No errors.
        linkable_spec_set = group_specs_by_type(group_by_item_specs)
        logger.debug(
            LazyFormat(lambda: f"Group-by-items were resolved to:\n{mf_pformat(linkable_spec_set.linkable_specs)}")
        )

        # Run post-resolution validation rules to generate issues that are generated at the query-level.
        query_level_issue_set = self._post_resolution_query_validator.validate_query(
            resolution_dag=resolution_dag,
            resolver_input_for_query=resolver_input_for_query,
            validation_rules=(
                MetricTimeQueryValidationRule(
                    self._manifest_lookup, resolver_input_for_query, resolve_group_by_item_result
                ),
                DuplicateMetricValidationRule(
                    self._manifest_lookup, resolver_input_for_query, resolve_group_by_item_result
                ),
            ),
        )

        if query_level_issue_set.has_issues:
            mappings_to_merge.append(
                InputToIssueSetMapping(
                    (
                        InputToIssueSetMappingItem(
                            resolver_input=resolver_input_for_query,
                            issue_set=query_level_issue_set,
                        ),
                    )
                )
            )

        issue_set_mapping = InputToIssueSetMapping.merge_iterable(mappings_to_merge)

        if issue_set_mapping.has_issues:
            return MetricFlowQueryResolution(
                query_spec=None,
                resolution_dag=resolution_dag,
                filter_spec_lookup=filter_spec_lookup,
                input_to_issue_set=issue_set_mapping,
                queried_semantic_models=(),
            )

        model_reference_set = set(resolve_group_by_item_result.linkable_element_set.derived_from_semantic_models)
        for filter_spec_resolution in filter_spec_lookup.spec_resolutions:
            model_reference_set.update(
                set(filter_spec_resolution.resolved_linkable_element_set.derived_from_semantic_models)
            )

        # Collect all semantic models referenced by the query.
        semantic_models_in_group_by_items = set(
            resolve_group_by_item_result.linkable_element_set.derived_from_semantic_models
        )
        semantic_models_in_filters = set(
            itertools.chain.from_iterable(
                filter_spec_resolution.resolved_linkable_element_set.derived_from_semantic_models
                for filter_spec_resolution in filter_spec_lookup.spec_resolutions
            )
        )
        measure_semantic_models = self._get_models_for_measures(resolution_dag)

        queried_semantic_models = set.union(
            semantic_models_in_group_by_items, semantic_models_in_filters, measure_semantic_models
        )
        queried_semantic_models -= {SemanticModelDerivation.VIRTUAL_SEMANTIC_MODEL_REFERENCE}

        # Sanity check to make sure that all queried semantic models are in the model.
        models_not_in_manifest = queried_semantic_models - {
            semantic_model.reference for semantic_model in self._manifest_lookup.semantic_manifest.semantic_models
        }

        # There are no known cases where this should happen, but adding this check just in case there's a bug where
        # a measure alias is used incorrectly.
        if len(models_not_in_manifest) > 0:
            logger.error(
                LazyFormat(
                    lambda: mf_pformat_dict(
                        "Semantic references that aren't in the manifest were found in the set used in "
                        "a query. This is a bug, and to avoid potential issues, they will be filtered out.",
                        {"models_not_in_manifest": models_not_in_manifest},
                    )
                )
            )
        queried_semantic_models -= models_not_in_manifest

        return MetricFlowQueryResolution(
            query_spec=MetricFlowQuerySpec(
                metric_specs=metric_specs,
                dimension_specs=linkable_spec_set.dimension_specs,
                entity_specs=linkable_spec_set.entity_specs,
                time_dimension_specs=linkable_spec_set.time_dimension_specs,
                order_by_specs=tuple(order_by_specs),
                limit=limit_input.limit,
                filter_intersection=query_level_filter_input.where_filter_intersection,
                filter_spec_resolution_lookup=filter_spec_lookup,
                min_max_only=min_max_only_input.min_max_only,
            ),
            resolution_dag=resolution_dag,
            filter_spec_lookup=filter_spec_lookup,
            input_to_issue_set=issue_set_mapping,
            queried_semantic_models=sorted_semantic_model_references(queried_semantic_models),
        )

    def _get_models_for_measures(self, resolution_dag: GroupByItemResolutionDag) -> Set[SemanticModelReference]:
        """Return the semantic model references for the measures used in the query."""
        resolution_dag_node_set = resolution_dag.sink_node.inclusive_ancestors()

        measure_references: Set[MeasureReference] = set()

        # Collect measures for metrics through the associated measure nodes.
        for measure_node in resolution_dag_node_set.measure_nodes:
            measure_references.add(measure_node.measure_reference)

        # For conversion metrics, get the measures through the metric since those measures aren't in the DAG.
        for metric_node in resolution_dag_node_set.metric_nodes:
            metric = self._manifest_lookup.metric_lookup.get_metric(metric_node.metric_reference)
            conversion_type_params = metric.type_params.conversion_type_params
            if conversion_type_params is None:
                continue

            # The base measure should be in a DAG, but just in case.
            measure_references.add(conversion_type_params.base_measure.measure_reference)
            measure_references.add(conversion_type_params.conversion_measure.measure_reference)

        model_references: Set[SemanticModelReference] = set()
        for measure_reference in measure_references:
            measure_semantic_model = self._manifest_lookup.semantic_model_lookup.measure_lookup.get_properties(
                measure_reference
            ).model_reference
            model_references.add(measure_semantic_model)

        return model_references



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True, eq=False)
class MinMaxNode(DataflowPlanNode):
    """Calculate the min and max of a single instance data set."""

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 1

    @staticmethod
    def create(parent_node: DataflowPlanNode) -> MinMaxNode:  # noqa: D102
        return MinMaxNode(parent_nodes=(parent_node,))

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_MIN_MAX_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_min_max_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return "Calculate min and max"

    @property
    def parent_node(self) -> DataflowPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return isinstance(other_node, self.__class__)

    def with_new_parents(self, new_parent_nodes: Sequence[DataflowPlanNode]) -> MinMaxNode:  # noqa: D102
        assert len(new_parent_nodes) == 1
        return MinMaxNode.create(parent_node=new_parent_nodes[0])



from __future__ import annotations

import textwrap

from dbt_semantic_interfaces.implementations.project_configuration import (
    PydanticProjectConfiguration,
)
from dbt_semantic_interfaces.implementations.time_spine_table_configuration import (
    PydanticTimeSpineTableConfiguration,
)
from dbt_semantic_interfaces.parsing.objects import YamlConfigFile
from dbt_semantic_interfaces.type_enums import TimeGranularity

EXAMPLE_PROJECT_CONFIGURATION = PydanticProjectConfiguration(
    time_spine_table_configurations=[
        PydanticTimeSpineTableConfiguration(
            location="example_schema.example_table",
            column_name="ds",
            grain=TimeGranularity.DAY,
        )
    ],
)

EXAMPLE_PROJECT_CONFIGURATION_YAML_CONFIG_FILE = YamlConfigFile(
    filepath="projection_configuration_yaml_file_path",
    contents=textwrap.dedent(
        """\
        project_configuration:
          time_spine_table_configurations:
            - location: example_schema.example_table
              column_name: ds
              grain: day
        """
    ),
)



from __future__ import annotations

import itertools
import logging
from collections import defaultdict
from typing import Dict, List, Sequence, Set

from dbt_semantic_interfaces.call_parameter_sets import FilterCallParameterSets, MetricCallParameterSet
from dbt_semantic_interfaces.implementations.filters.where_filter import PydanticWhereFilterIntersection
from dbt_semantic_interfaces.protocols import WhereFilter
from dbt_semantic_interfaces.references import EntityReference
from typing_extensions import override

from metricflow_semantics.mf_logging.runtime import log_runtime
from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.naming.object_builder_str import ObjectBuilderNameConverter
from metricflow_semantics.query.group_by_item.candidate_push_down.push_down_visitor import DagTraversalPathTracker
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_location import (
    WhereFilterLocation,
)
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_pattern_factory import (
    WhereFilterPatternFactory,
)
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_spec_lookup import (
    FilterSpecResolution,
    FilterSpecResolutionLookUp,
    NonParsableFilterResolution,
    PatternAssociationForWhereFilterGroupByItem,
    ResolvedSpecLookUpKey,
)
from metricflow_semantics.query.group_by_item.group_by_item_resolver import GroupByItemResolver
from metricflow_semantics.query.group_by_item.resolution_dag.dag import GroupByItemResolutionDag, ResolutionDagSinkNode
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.base_node import (
    GroupByItemResolutionNodeVisitor,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.measure_source_node import (
    MeasureGroupByItemSourceNode,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.metric_resolution_node import (
    MetricGroupByItemResolutionNode,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.no_metrics_query_source_node import (
    NoMetricsGroupByItemSourceNode,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.query_resolution_node import (
    QueryGroupByItemResolutionNode,
)
from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.filter_spec_resolver.invalid_where import WhereFilterParsingIssue
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryResolutionIssueSet,
)

logger = logging.getLogger(__name__)


class WhereFilterSpecResolver:
    """Resolves the specs for the (ambiguous) group-by-items that are in where filters of the resolution DAG.

    In the resolution DAG, filters can occur:

    * In the query.
    * In the definition of a metric as a filter.
    * In the definition of an input measure to a base metric.
    * In the definition of an input metric to a derived metric.

    The concrete specs for the group-by-items are returned in a lookup.
    """

    def __init__(  # noqa: D107
        self,
        manifest_lookup: SemanticManifestLookup,
        resolution_dag: GroupByItemResolutionDag,
        spec_pattern_factory: WhereFilterPatternFactory,
    ) -> None:
        self._manifest_lookup = manifest_lookup
        self._resolution_dag = resolution_dag
        self.spec_pattern_factory = spec_pattern_factory

    def resolve_lookup(self) -> FilterSpecResolutionLookUp:
        """Find all where filters and return a lookup that provides the specs for the included group-by-items."""
        # Workaround for a Pycharm type inspection issue with decorators.
        # noinspection PyArgumentList
        return self._resolve_lookup()

    @log_runtime()
    def _resolve_lookup(self) -> FilterSpecResolutionLookUp:
        visitor = _ResolveWhereFilterSpecVisitor(
            manifest_lookup=self._manifest_lookup,
            spec_pattern_factory=self.spec_pattern_factory,
        )
        return self._resolution_dag.sink_node.accept(visitor)


class _ResolveWhereFilterSpecVisitor(GroupByItemResolutionNodeVisitor[FilterSpecResolutionLookUp]):
    """Visitor that resolves specs for all filters in a resolution DAG and builds a lookup.

    This visitor traverses the group-by-item resolution DAG and if there is a where filter associated with the node,
    this resolves (ambiguous) group-by-items that are specified in the filter. All resolutions are
    collected and returned in a lookup object.
    """

    def __init__(
        self, manifest_lookup: SemanticManifestLookup, spec_pattern_factory: WhereFilterPatternFactory
    ) -> None:
        self._manifest_lookup = manifest_lookup
        self._path_from_start_node_tracker = DagTraversalPathTracker()
        self._spec_pattern_factory = spec_pattern_factory

    @staticmethod
    def _dedupe_filter_call_parameter_sets(
        filter_call_parameter_sets_sequence: Sequence[FilterCallParameterSets],
    ) -> FilterCallParameterSets:
        # FilterCallParameterSets needs an update.
        return FilterCallParameterSets(
            dimension_call_parameter_sets=tuple(
                dict.fromkeys(
                    itertools.chain.from_iterable(
                        filter_call_parameter_sets.dimension_call_parameter_sets
                        for filter_call_parameter_sets in filter_call_parameter_sets_sequence
                    )
                )
            ),
            time_dimension_call_parameter_sets=tuple(
                dict.fromkeys(
                    itertools.chain.from_iterable(
                        filter_call_parameter_sets.time_dimension_call_parameter_sets
                        for filter_call_parameter_sets in filter_call_parameter_sets_sequence
                    )
                )
            ),
            entity_call_parameter_sets=tuple(
                dict.fromkeys(
                    itertools.chain.from_iterable(
                        filter_call_parameter_sets.entity_call_parameter_sets
                        for filter_call_parameter_sets in filter_call_parameter_sets_sequence
                    )
                )
            ),
            metric_call_parameter_sets=tuple(
                dict.fromkeys(
                    itertools.chain.from_iterable(
                        filter_call_parameter_sets.metric_call_parameter_sets
                        for filter_call_parameter_sets in filter_call_parameter_sets_sequence
                    )
                )
            ),
        )

    def _map_filter_parameter_sets_to_pattern(
        self,
        filter_call_parameter_sets: FilterCallParameterSets,
    ) -> Sequence[PatternAssociationForWhereFilterGroupByItem]:
        """Given the call parameter sets in a filter, map them to spec patterns.

        If a given call parameter set has already been resolved and in spec_resolution_lookup_so_far, it is skipped and
        not returned in the dictionary.

        This assumes that the items in filter_call_parameter_sets have been deduped.
        """
        patterns_in_filter: List[PatternAssociationForWhereFilterGroupByItem] = []
        for dimension_call_parameter_set in filter_call_parameter_sets.dimension_call_parameter_sets:
            patterns_in_filter.append(
                PatternAssociationForWhereFilterGroupByItem(
                    call_parameter_set=dimension_call_parameter_set,
                    object_builder_str=ObjectBuilderNameConverter.input_str_from_dimension_call_parameter_set(
                        dimension_call_parameter_set
                    ),
                    spec_pattern=self._spec_pattern_factory.create_for_dimension_call_parameter_set(
                        dimension_call_parameter_set
                    ),
                )
            )
        for time_dimension_call_parameter_set in filter_call_parameter_sets.time_dimension_call_parameter_sets:
            patterns_in_filter.append(
                PatternAssociationForWhereFilterGroupByItem(
                    call_parameter_set=time_dimension_call_parameter_set,
                    object_builder_str=ObjectBuilderNameConverter.input_str_from_time_dimension_call_parameter_set(
                        time_dimension_call_parameter_set
                    ),
                    spec_pattern=self._spec_pattern_factory.create_for_time_dimension_call_parameter_set(
                        time_dimension_call_parameter_set
                    ),
                )
            )
        for entity_call_parameter_set in filter_call_parameter_sets.entity_call_parameter_sets:
            patterns_in_filter.append(
                PatternAssociationForWhereFilterGroupByItem(
                    call_parameter_set=entity_call_parameter_set,
                    object_builder_str=ObjectBuilderNameConverter.input_str_from_entity_call_parameter_set(
                        entity_call_parameter_set
                    ),
                    spec_pattern=self._spec_pattern_factory.create_for_entity_call_parameter_set(
                        entity_call_parameter_set
                    ),
                )
            )
        for metric_call_parameter_set in filter_call_parameter_sets.metric_call_parameter_sets:
            # Convert LinkableElementReferences to EntityReferences. Will need to support dimensions later.
            updated_metric_call_parameter_set = MetricCallParameterSet(
                metric_reference=metric_call_parameter_set.metric_reference,
                group_by=tuple(
                    EntityReference(element_name=group_by_ref.element_name)
                    for group_by_ref in metric_call_parameter_set.group_by
                ),
            )
            patterns_in_filter.append(
                PatternAssociationForWhereFilterGroupByItem(
                    call_parameter_set=updated_metric_call_parameter_set,
                    object_builder_str=ObjectBuilderNameConverter.input_str_from_metric_call_parameter_set(
                        metric_call_parameter_set
                    ),
                    spec_pattern=self._spec_pattern_factory.create_for_metric_call_parameter_set(
                        metric_call_parameter_set
                    ),
                )
            )

        return patterns_in_filter

    @override
    def visit_measure_node(self, node: MeasureGroupByItemSourceNode) -> FilterSpecResolutionLookUp:
        """No filters are defined in measures, so this is a no-op."""
        with self._path_from_start_node_tracker.track_node_visit(node):
            return FilterSpecResolutionLookUp.empty_instance()

    @override
    def visit_metric_node(self, node: MetricGroupByItemResolutionNode) -> FilterSpecResolutionLookUp:
        """Resolve specs for filters in a metric definition."""
        with self._path_from_start_node_tracker.track_node_visit(node) as resolution_path:
            results_to_merge: List[FilterSpecResolutionLookUp] = []
            for parent_node in node.parent_nodes:
                results_to_merge.append(parent_node.accept(self))

            resolved_spec_lookup_so_far = FilterSpecResolutionLookUp.merge_iterable(results_to_merge)

            return resolved_spec_lookup_so_far.merge(
                self._resolve_specs_for_where_filters(
                    current_node=node,
                    resolution_path=resolution_path,
                    where_filters_and_locations=self._get_where_filters_at_metric_node(node),
                )
            )

    @override
    def visit_query_node(self, node: QueryGroupByItemResolutionNode) -> FilterSpecResolutionLookUp:
        """Resolve specs for filters defined for the query."""
        with self._path_from_start_node_tracker.track_node_visit(node) as resolution_path:
            results_to_merge: List[FilterSpecResolutionLookUp] = []
            for parent_node in node.parent_nodes:
                results_to_merge.append(parent_node.accept(self))

            # If the same metric is present multiple times in a query - there could be duplicates.
            resolved_spec_lookup_so_far = FilterSpecResolutionLookUp.merge_iterable(results_to_merge)

            return resolved_spec_lookup_so_far.merge(
                self._resolve_specs_for_where_filters(
                    current_node=node,
                    resolution_path=resolution_path,
                    where_filters_and_locations={
                        WhereFilterLocation.for_query(node.metrics_in_query): set(
                            node.where_filter_intersection.where_filters
                        )
                    },
                )
            )

    @override
    def visit_no_metrics_query_node(self, node: NoMetricsGroupByItemSourceNode) -> FilterSpecResolutionLookUp:
        """Similar to the measure node - filters are applied at the query level."""
        with self._path_from_start_node_tracker.track_node_visit(node):
            return FilterSpecResolutionLookUp.empty_instance()

    def _get_where_filters_at_metric_node(
        self, metric_node: MetricGroupByItemResolutionNode
    ) -> Dict[WhereFilterLocation, Set[WhereFilter]]:
        """Return the filters used with a metric in the manifest.

        A derived metric definition can have a filter for an input metric, and we'll need to resolve that when the
        dataflow plan is built.
        """
        where_filters_and_locations: Dict[WhereFilterLocation, Set[WhereFilter]] = defaultdict(set)
        metric = self._manifest_lookup.metric_lookup.get_metric(metric_node.metric_reference)

        metric_filter_location = WhereFilterLocation.for_metric(metric_reference=metric_node.metric_reference)
        if metric.input_measures is not None:
            for input_measure in metric.input_measures:
                if input_measure.filter is not None:
                    for input_measure_where_filter in input_measure.filter.where_filters:
                        where_filters_and_locations[metric_filter_location].add(input_measure_where_filter)

        if metric.filter is not None:
            for metric_where_filter in metric.filter.where_filters:
                where_filters_and_locations[metric_filter_location].add(metric_where_filter)

        # This is a metric that is an input metric for a derived metric. The derived metric is the child node of this
        # metric node.
        if metric_node.metric_input_location is not None:
            child_metric_input = metric_node.metric_input_location.get_metric_input(self._manifest_lookup.metric_lookup)
            if child_metric_input.filter is not None:
                input_metric_filter_location = WhereFilterLocation.for_input_metric(
                    input_metric_reference=metric_node.metric_reference
                )
                for input_metric_where_filter in child_metric_input.filter.where_filters:
                    where_filters_and_locations[input_metric_filter_location].add(input_metric_where_filter)

        return where_filters_and_locations

    def _resolve_specs_for_where_filters(
        self,
        current_node: ResolutionDagSinkNode,
        resolution_path: MetricFlowQueryResolutionPath,
        where_filters_and_locations: Dict[WhereFilterLocation, Set[WhereFilter]],
    ) -> FilterSpecResolutionLookUp:
        """Given the filters at a particular node, build the spec lookup for those filters.

        The start node should be the query node.
        """
        resolution_dag = GroupByItemResolutionDag(
            sink_node=current_node,
        )
        group_by_item_resolver = GroupByItemResolver(
            manifest_lookup=self._manifest_lookup,
            resolution_dag=resolution_dag,
        )
        non_parsable_resolutions: List[NonParsableFilterResolution] = []
        filter_call_parameter_sets_by_location: Dict[WhereFilterLocation, List[FilterCallParameterSets]] = defaultdict(
            list
        )
        # No input metric in locations when we get here
        for location, where_filters in where_filters_and_locations.items():
            for where_filter in where_filters:
                try:
                    filter_call_parameter_sets = where_filter.call_parameter_sets
                except Exception as e:
                    non_parsable_resolutions.append(
                        NonParsableFilterResolution(
                            filter_location_path=resolution_path,
                            where_filter_intersection=PydanticWhereFilterIntersection(
                                where_filters=list(
                                    {
                                        where_filter
                                        for where_filter_set in where_filters_and_locations.values()
                                        for where_filter in where_filter_set
                                    }
                                )
                            ),
                            issue_set=MetricFlowQueryResolutionIssueSet.from_issue(
                                WhereFilterParsingIssue.from_parameters(
                                    where_filter=where_filter,
                                    parse_exception=e,
                                    query_resolution_path=resolution_path,
                                )
                            ),
                        )
                    )
                    continue
                filter_call_parameter_sets_by_location[location].append(filter_call_parameter_sets)

        deduped_filter_call_parameter_sets_by_location = {
            location: _ResolveWhereFilterSpecVisitor._dedupe_filter_call_parameter_sets(filter_call_parameter_sets_list)
            for location, filter_call_parameter_sets_list in filter_call_parameter_sets_by_location.items()
        }

        resolutions: List[FilterSpecResolution] = []
        for (
            filter_location,
            deduped_filter_call_parameter_sets,
        ) in deduped_filter_call_parameter_sets_by_location.items():
            for group_by_item_in_where_filter in self._map_filter_parameter_sets_to_pattern(
                filter_call_parameter_sets=deduped_filter_call_parameter_sets,
            ):
                group_by_item_resolution = group_by_item_resolver.resolve_matching_item_for_filters(
                    input_str=group_by_item_in_where_filter.object_builder_str,
                    spec_pattern=group_by_item_in_where_filter.spec_pattern,
                    resolution_node=current_node,
                    filter_location=filter_location,
                )
                # The paths in the issue set are generated relative to the current node. For error messaging, it seems more
                # helpful for those paths to be relative to the query. To do, we have to add nodes from the resolution path.
                # e.g. if the current node is B, and the resolution path is [A, B], an issue might have the relative path
                # [B, C]. To join those paths to produce [A, B, C], the path prefix should be [A].
                path_prefix = MetricFlowQueryResolutionPath(
                    resolution_path_nodes=resolution_path.resolution_path_nodes[:-1]
                )
                resolutions.append(
                    FilterSpecResolution(
                        lookup_key=ResolvedSpecLookUpKey(
                            filter_location=filter_location,
                            call_parameter_set=group_by_item_in_where_filter.call_parameter_set,
                        ),
                        filter_location_path=resolution_path,
                        resolved_linkable_element_set=group_by_item_resolution.linkable_element_set,
                        where_filter_intersection=PydanticWhereFilterIntersection(
                            where_filters=list(where_filters_and_locations[filter_location])
                        ),
                        spec_pattern=group_by_item_in_where_filter.spec_pattern,
                        issue_set=group_by_item_resolution.issue_set.with_path_prefix(path_prefix),
                        object_builder_str=group_by_item_in_where_filter.object_builder_str,
                    )
                )

        if any(resolution.issue_set.has_errors for resolution in resolutions) or len(non_parsable_resolutions) > 0:
            return FilterSpecResolutionLookUp(
                spec_resolutions=tuple(resolution for resolution in resolutions if resolution.issue_set.has_errors),
                non_parsable_resolutions=tuple(non_parsable_resolutions),
            )

        return FilterSpecResolutionLookUp(
            spec_resolutions=tuple(resolutions),
            non_parsable_resolutions=tuple(non_parsable_resolutions),
        )



from __future__ import annotations

import logging
from collections import defaultdict
from typing import Dict, FrozenSet, List, Set, Tuple

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.sql.sql_exprs import SqlExpressionTreeLineage
from typing_extensions import override

from metricflow.sql.optimizer.cte_alias_to_cte_node_mapping import SqlCteAliasMappingLookup
from metricflow.sql.optimizer.tag_column_aliases import NodeToColumnAliasMapping
from metricflow.sql.sql_plan import (
    SqlCreateTableAsNode,
    SqlCteAliasMapping,
    SqlCteNode,
    SqlPlanNode,
    SqlPlanNodeVisitor,
    SqlSelectColumn,
    SqlSelectQueryFromClauseNode,
    SqlSelectStatementNode,
    SqlTableNode,
)

logger = logging.getLogger(__name__)


class SqlMapRequiredColumnAliasesVisitor(SqlPlanNodeVisitor[None]):
    """To aid column pruning, traverse the SQL-query representation DAG and map the SELECT columns needed at each node.

    For example, the query:

        -- SELECT node_id="select_0"
        SELECT source_0.col_0 AS col_0_renamed
        FROM (
            -- SELECT node_id="select_1
            SELECT
                example_table.col_0
                example_table.col_1
            FROM example_table_0
        ) source_0

    would generate the mapping:

        {
            "select_0": {"col_0"},
            "select_1": {"col_0"),
        }

    The mapping can be later used to rewrite the query to:

        SELECT source_0.col_0 AS col_0_renamed
        FROM (
            SELECT
                example_table.col_0
            FROM example_table_0
        ) source_0
    """

    def __init__(
        self,
        start_node: SqlPlanNode,
        required_column_aliases_in_start_node: FrozenSet[str],
        cte_alias_mapping_lookup: SqlCteAliasMappingLookup,
    ) -> None:
        """Initializer.

        Args:
            start_node: The node where the traversal by this visitor will start.
            required_column_aliases_in_start_node: The column aliases at the `start_node` that are required.
        """
        # Stores the mapping of the SQL node to the required column aliases. This will be updated as the visitor
        # traverses the SQL-query representation DAG.
        self._current_required_column_alias_mapping = NodeToColumnAliasMapping()
        self._current_required_column_alias_mapping.add_aliases(start_node, required_column_aliases_in_start_node)

        # Helps lookup the CTE node associated with a given CTE alias. A member variable is needed as any node in the
        # SQL DAG can reference a CTE.
        self._cte_node_lookup = cte_alias_mapping_lookup

    def _search_for_expressions(
        self, select_node: SqlSelectStatementNode, pruned_select_columns: Tuple[SqlSelectColumn, ...]
    ) -> SqlExpressionTreeLineage:
        """Returns the expressions used in the immediate select statement.

        i.e. this does not return expressions used in sub-queries. pruned_select_columns needs to be passed in since the
        node may have the select columns pruned.
        """
        all_expr_search_results: List[SqlExpressionTreeLineage] = []

        for select_column in pruned_select_columns:
            all_expr_search_results.append(select_column.expr.lineage)

        for join_description in select_node.join_descs:
            if join_description.on_condition:
                all_expr_search_results.append(join_description.on_condition.lineage)

        for group_by in select_node.group_bys:
            all_expr_search_results.append(group_by.expr.lineage)

        for order_by in select_node.order_bys:
            all_expr_search_results.append(order_by.expr.lineage)

        if select_node.where:
            all_expr_search_results.append(select_node.where.lineage)

        return SqlExpressionTreeLineage.merge_iterable(all_expr_search_results)

    @override
    def visit_cte_node(self, node: SqlCteNode) -> None:
        select_statement = node.select_statement
        # Copy the tagged aliases from the CTE to the SELECT since when visiting a SELECT, the CTE node (not the SELECT
        # in the CTE) was tagged with the required aliases.
        required_column_aliases_in_this_node = self._current_required_column_alias_mapping.get_aliases(node)
        self._current_required_column_alias_mapping.add_aliases(select_statement, required_column_aliases_in_this_node)
        # Visit parent nodes.
        select_statement.accept(self)

    def _visit_parents(self, node: SqlPlanNode) -> None:
        """Default recursive handler to visit the parents of the given node."""
        for parent_node in node.parent_nodes:
            parent_node.accept(self)
        return

    def _tag_potential_cte_node(
        self, cte_alias_mapping: SqlCteAliasMapping, table_name: str, column_aliases: Set[str]
    ) -> None:
        """A reference to a SQL table might be a CTE. If so, tag the appropriate aliases in the CTEs."""
        cte_node = cte_alias_mapping.get_cte_node_for_alias(table_name)

        if cte_node is not None:
            self._current_required_column_alias_mapping.add_aliases(cte_node, column_aliases)
            # `visit_cte_node` will handle propagating the required aliases to all CTEs that this CTE node depends on.
            cte_node.accept(self)

    def visit_select_statement_node(self, node: SqlSelectStatementNode) -> None:
        """Based on required column aliases for this SELECT, figure out required column aliases in parents."""
        cte_alias_mapping = self._cte_node_lookup.get_cte_alias_mapping(node)

        initial_required_column_aliases_in_this_node = self._current_required_column_alias_mapping.get_aliases(node)

        # If this SELECT statement uses DISTINCT, all columns are required as removing them would change the meaning of
        # the query.
        updated_required_column_aliases_in_this_node = set(initial_required_column_aliases_in_this_node)
        if node.distinct:
            updated_required_column_aliases_in_this_node.update(
                {select_column.column_alias for select_column in node.select_columns}
            )

        # Any columns in the group by also need to be kept to have a correct query.
        updated_required_column_aliases_in_this_node.update(
            {group_by_select_column.column_alias for group_by_select_column in node.group_bys}
        )
        logger.debug(
            LazyFormat(
                "Tagging column aliases in parent nodes given what's required in this node",
                this_node=node,
                initial_required_column_aliases_in_this_node=list(initial_required_column_aliases_in_this_node),
                updated_required_column_aliases_in_this_node=list(updated_required_column_aliases_in_this_node),
            )
        )
        # Since additional select columns could have been selected due to DISTINCT or GROUP BY, re-tag.
        self._current_required_column_alias_mapping.add_aliases(node, updated_required_column_aliases_in_this_node)

        required_select_columns_in_this_node = tuple(
            select_column
            for select_column in node.select_columns
            if select_column.column_alias in updated_required_column_aliases_in_this_node
        )

        if len(required_select_columns_in_this_node) == 0:
            raise RuntimeError(
                "No columns are required in this node - this indicates a bug in this visitor or in the inputs."
            )
        # It's possible for `required_select_columns_in_this_node` to be empty because we traverse through the ancestors
        # of a CTE node whenever a CTE node is updated. See `test_multi_child_pruning`.

        # Based on the expressions in this select statement, figure out what column aliases are needed in the sources of
        # this query (i.e. tables or sub-queries in the FROM or JOIN clauses).
        exprs_used_in_this_node = self._search_for_expressions(node, required_select_columns_in_this_node)

        # If any of the string expressions don't have context on what columns are used in the expression, then it's
        # impossible to know what columns can be pruned from the parent sources. Tag all columns in parents as required.
        if any([string_expr.used_columns is None for string_expr in exprs_used_in_this_node.string_exprs]):
            nodes_to_retain_all_columns = [node.from_source]
            for join_desc in node.join_descs:
                nodes_to_retain_all_columns.append(join_desc.right_source)

            for node_to_retain_all_columns in nodes_to_retain_all_columns:
                nearest_select_columns = node_to_retain_all_columns.nearest_select_columns(cte_alias_mapping)
                for select_column in nearest_select_columns or ():
                    self._current_required_column_alias_mapping.add_alias(
                        node=node_to_retain_all_columns, column_alias=select_column.column_alias
                    )
            # TODO: TBD - may be necessary to mark columns in all visible CTEs since a string can reference anything.
            self._visit_parents(node)
            return

        # Create a mapping from the source alias to the column aliases needed from the corresponding source.
        source_alias_to_required_column_aliases: Dict[str, Set[str]] = defaultdict(set)
        for column_reference_expr in exprs_used_in_this_node.column_reference_exprs:
            column_reference = column_reference_expr.col_ref
            source_alias_to_required_column_aliases[column_reference.table_alias].add(column_reference.column_name)

        logger.debug(
            LazyFormat(
                "Collected required column names from sources",
                source_alias_to_required_column_aliases=source_alias_to_required_column_aliases,
            )
        )
        # Appropriately tag the columns required in the parent nodes.
        if node.from_source_alias in source_alias_to_required_column_aliases:
            aliases_required_in_parent = source_alias_to_required_column_aliases[node.from_source_alias]
            self._current_required_column_alias_mapping.add_aliases(
                node=node.from_source, column_aliases=aliases_required_in_parent
            )
            from_source_as_sql_table_node = node.from_source.as_sql_table_node
            if from_source_as_sql_table_node is not None:
                self._tag_potential_cte_node(
                    cte_alias_mapping=cte_alias_mapping,
                    table_name=from_source_as_sql_table_node.sql_table.table_name,
                    column_aliases=aliases_required_in_parent,
                )
        for join_desc in node.join_descs:
            if join_desc.right_source_alias in source_alias_to_required_column_aliases:
                aliases_required_in_parent = source_alias_to_required_column_aliases[join_desc.right_source_alias]
                self._current_required_column_alias_mapping.add_aliases(
                    node=join_desc.right_source, column_aliases=aliases_required_in_parent
                )
                right_source_as_sql_table_node = join_desc.right_source.as_sql_table_node
                if right_source_as_sql_table_node is not None:
                    self._tag_potential_cte_node(
                        cte_alias_mapping=cte_alias_mapping,
                        table_name=right_source_as_sql_table_node.sql_table.table_name,
                        column_aliases=aliases_required_in_parent,
                    )

        # For all string columns, assume that they are needed from all sources since we don't have a table alias
        # in SqlStringExpression.used_columns
        for string_expr in exprs_used_in_this_node.string_exprs:
            if string_expr.used_columns:
                for column_alias in string_expr.used_columns:
                    for node_to_retain_all_columns in (node.from_source,) + tuple(
                        join_desc.right_source for join_desc in node.join_descs
                    ):
                        self._current_required_column_alias_mapping.add_alias(node_to_retain_all_columns, column_alias)

        # Same with unqualified column references - it's hard to tell which source it came from, so it's safest to say
        # it's required from all parents.
        # An unqualified column reference expression is like `SELECT col_0` whereas a qualified column reference
        # expression is like `SELECT table_0.col_0`.
        for unqualified_column_reference_expr in exprs_used_in_this_node.column_alias_reference_exprs:
            column_alias = unqualified_column_reference_expr.column_alias
            for node_to_retain_all_columns in (node.from_source,) + tuple(
                join_desc.right_source for join_desc in node.join_descs
            ):
                self._current_required_column_alias_mapping.add_alias(node_to_retain_all_columns, column_alias)

        # Visit recursively.
        self._visit_parents(node)
        return

    def visit_table_node(self, node: SqlTableNode) -> None:
        """There are no SELECT columns in this node, so pruning cannot apply."""
        return

    def visit_query_from_clause_node(self, node: SqlSelectQueryFromClauseNode) -> None:
        """Pruning cannot be done here since this is an arbitrary user-provided SQL query."""
        return

    def visit_create_table_as_node(self, node: SqlCreateTableAsNode) -> None:  # noqa: D102
        return self._visit_parents(node)

    @property
    def required_column_alias_mapping(self) -> NodeToColumnAliasMapping:
        """Return the column aliases required at each node as determined after traversal."""
        return self._current_required_column_alias_mapping



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, Optional, Sequence, Set, Tuple

from dbt_semantic_interfaces.type_enums import TimeGranularity
from typing_extensions import override

from metricflow_semantics.specs.instance_spec import InstanceSpec, LinkableInstanceSpec
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern
from metricflow_semantics.specs.spec_set import group_specs_by_type
from metricflow_semantics.specs.time_dimension_spec import (
    DEFAULT_TIME_GRANULARITY,
    TimeDimensionSpec,
    TimeDimensionSpecComparisonKey,
    TimeDimensionSpecField,
)
from metricflow_semantics.time.granularity import ExpandedTimeGranularity


@dataclass(frozen=True)
class MetricTimeDefaultGranularityPattern(SpecPattern):
    """A pattern that matches metric_time specs if they have the default granularity for the requested metrics.

    This is used to determine the granularity that should be used for metric_time if no granularity is specified.
    Spec passes through if granularity is already selected or if no metrics were queried, since no default is needed.
    All non-metric_time specs are passed through.

    e.g., if a metric with default_granularity MONTH is queried

    inputs:
        [
            TimeDimensionSpec('metric_time', 'day'),
            TimeDimensionSpec('metric_time', 'week'),
            TimeDimensionSpec('metric_time', 'month'),
            DimensionSpec('listing__country'),
        ]

    matches:
        [
            TimeDimensionSpec('metric_time', 'month'),
            DimensionSpec('listing__country'),
        ]
    """

    max_metric_default_time_granularity: Optional[TimeGranularity]

    @override
    def match(self, candidate_specs: Sequence[InstanceSpec]) -> Sequence[InstanceSpec]:
        spec_set = group_specs_by_type(candidate_specs)
        # If there are no metric_time specs in the query, skip this filter.
        if not spec_set.metric_time_specs:
            return candidate_specs

        # If there are metrics in the query, use max metric default. For no-metric queries, use standard default.
        # TODO: [custom granularity] allow custom granularities to be used as defaults if appropriate
        default_granularity = ExpandedTimeGranularity.from_time_granularity(
            self.max_metric_default_time_granularity or DEFAULT_TIME_GRANULARITY
        )

        spec_key_to_grains: Dict[TimeDimensionSpecComparisonKey, Set[ExpandedTimeGranularity]] = defaultdict(set)
        spec_key_to_specs: Dict[TimeDimensionSpecComparisonKey, Tuple[TimeDimensionSpec, ...]] = defaultdict(tuple)
        for metric_time_spec in spec_set.metric_time_specs:
            spec_key = metric_time_spec.comparison_key(exclude_fields=(TimeDimensionSpecField.TIME_GRANULARITY,))
            spec_key_to_grains[spec_key].add(metric_time_spec.time_granularity)
            spec_key_to_specs[spec_key] += (metric_time_spec,)

        matched_metric_time_specs: Tuple[TimeDimensionSpec, ...] = ()
        for spec_key, time_grains in spec_key_to_grains.items():
            if default_granularity in time_grains:
                matched_metric_time_specs += (spec_key_to_specs[spec_key][0].with_grain(default_granularity),)
            else:
                # If default_granularity is not in the available options, then time granularity was specified in the request
                # and a default is not needed here. Pass all options through for this spec key.
                matched_metric_time_specs += spec_key_to_specs[spec_key]

        matching_specs: Sequence[LinkableInstanceSpec] = (
            spec_set.dimension_specs
            + matched_metric_time_specs
            + tuple(spec for spec in spec_set.time_dimension_specs if not spec.is_metric_time)
            + spec_set.entity_specs
            + spec_set.group_by_metric_specs
        )

        return matching_specs



from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Generic

from metricflow_semantics.visitor import VisitorOutputT
from typing_extensions import override

if typing.TYPE_CHECKING:
    from metricflow.dataflow.dataflow_plan import DataflowPlanNode
    from metricflow.dataflow.nodes.add_generated_uuid import AddGeneratedUuidColumnNode
    from metricflow.dataflow.nodes.aggregate_measures import AggregateMeasuresNode
    from metricflow.dataflow.nodes.alias_specs import AliasSpecsNode
    from metricflow.dataflow.nodes.combine_aggregated_outputs import CombineAggregatedOutputsNode
    from metricflow.dataflow.nodes.compute_metrics import ComputeMetricsNode
    from metricflow.dataflow.nodes.constrain_time import ConstrainTimeRangeNode
    from metricflow.dataflow.nodes.filter_elements import FilterElementsNode
    from metricflow.dataflow.nodes.join_conversion_events import JoinConversionEventsNode
    from metricflow.dataflow.nodes.join_over_time import JoinOverTimeRangeNode
    from metricflow.dataflow.nodes.join_to_base import JoinOnEntitiesNode
    from metricflow.dataflow.nodes.join_to_custom_granularity import JoinToCustomGranularityNode
    from metricflow.dataflow.nodes.join_to_time_spine import JoinToTimeSpineNode
    from metricflow.dataflow.nodes.metric_time_transform import MetricTimeDimensionTransformNode
    from metricflow.dataflow.nodes.min_max import MinMaxNode
    from metricflow.dataflow.nodes.order_by_limit import OrderByLimitNode
    from metricflow.dataflow.nodes.read_sql_source import ReadSqlSourceNode
    from metricflow.dataflow.nodes.semi_additive_join import SemiAdditiveJoinNode
    from metricflow.dataflow.nodes.where_filter import WhereConstraintNode
    from metricflow.dataflow.nodes.window_reaggregation_node import WindowReaggregationNode
    from metricflow.dataflow.nodes.write_to_data_table import WriteToResultDataTableNode
    from metricflow.dataflow.nodes.write_to_table import WriteToResultTableNode


class DataflowPlanNodeVisitor(Generic[VisitorOutputT], ABC):
    """An object that can be used to visit the nodes of a dataflow plan.

    Follows the visitor pattern: https://en.wikipedia.org/wiki/Visitor_pattern
    All visit* methods are similar and one exists for every type of node in the dataflow plan. The appropriate method
    will be called with DataflowPlanNode.accept().
    """

    @abstractmethod
    def visit_source_node(self, node: ReadSqlSourceNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_join_on_entities_node(self, node: JoinOnEntitiesNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_aggregate_measures_node(self, node: AggregateMeasuresNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_compute_metrics_node(self, node: ComputeMetricsNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_window_reaggregation_node(self, node: WindowReaggregationNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_order_by_limit_node(self, node: OrderByLimitNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_where_constraint_node(self, node: WhereConstraintNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_write_to_result_data_table_node(self, node: WriteToResultDataTableNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_write_to_result_table_node(self, node: WriteToResultTableNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_filter_elements_node(self, node: FilterElementsNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_combine_aggregated_outputs_node(self, node: CombineAggregatedOutputsNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_constrain_time_range_node(self, node: ConstrainTimeRangeNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_join_over_time_range_node(self, node: JoinOverTimeRangeNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_semi_additive_join_node(self, node: SemiAdditiveJoinNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_metric_time_dimension_transform_node(  # noqa: D102
        self, node: MetricTimeDimensionTransformNode
    ) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_join_to_time_spine_node(self, node: JoinToTimeSpineNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_min_max_node(self, node: MinMaxNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_add_generated_uuid_column_node(self, node: AddGeneratedUuidColumnNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_join_conversion_events_node(self, node: JoinConversionEventsNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_join_to_custom_granularity_node(self, node: JoinToCustomGranularityNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_alias_specs_node(self, node: AliasSpecsNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError


class DataflowPlanNodeVisitorWithDefaultHandler(DataflowPlanNodeVisitor[VisitorOutputT], Generic[VisitorOutputT]):
    """Similar to `DataflowPlanNodeVisitor`, but with an abstract default handler that gets called for each node.

    This could potentially replace `DataflowPlanNodeVisitor`.
    """

    @abstractmethod
    def _default_handler(self, node: DataflowPlanNode) -> VisitorOutputT:
        raise NotImplementedError

    @override
    def visit_source_node(self, node: ReadSqlSourceNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_join_on_entities_node(self, node: JoinOnEntitiesNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_aggregate_measures_node(self, node: AggregateMeasuresNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_compute_metrics_node(self, node: ComputeMetricsNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_window_reaggregation_node(self, node: WindowReaggregationNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_order_by_limit_node(self, node: OrderByLimitNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_where_constraint_node(self, node: WhereConstraintNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_write_to_result_data_table_node(self, node: WriteToResultDataTableNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_write_to_result_table_node(self, node: WriteToResultTableNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_filter_elements_node(self, node: FilterElementsNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_combine_aggregated_outputs_node(self, node: CombineAggregatedOutputsNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_constrain_time_range_node(self, node: ConstrainTimeRangeNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_join_over_time_range_node(self, node: JoinOverTimeRangeNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_semi_additive_join_node(self, node: SemiAdditiveJoinNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_metric_time_dimension_transform_node(  # noqa: D102
        self, node: MetricTimeDimensionTransformNode
    ) -> VisitorOutputT:
        return self._default_handler(node)

    @override
    def visit_join_to_time_spine_node(self, node: JoinToTimeSpineNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_min_max_node(self, node: MinMaxNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_add_generated_uuid_column_node(self, node: AddGeneratedUuidColumnNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_join_conversion_events_node(self, node: JoinConversionEventsNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_join_to_custom_granularity_node(self, node: JoinToCustomGranularityNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)

    @override
    def visit_alias_specs_node(self, node: AliasSpecsNode) -> VisitorOutputT:  # noqa: D102
        return self._default_handler(node)
